"""
This module contains wrappers for the parts making up a :mod:`swiftsimio`
dataset to facilitate analyses of individual simulated galaxies.

The top-level wrapper is :class:`SWIFTGalaxy`, which inherits from
:class:`~swiftsimio.reader.SWIFTDataset`. It extends the functionality of a
dataset to select particles belonging to a single galaxy, handle coordinate
transformations while keeping all particles in a consistent frame of reference,
providing spherical and cylindrical coordinates, and more.

Additional wrappers are provided for
:class:`swiftsimio.reader.__SWIFTGroupDataset` and
:class:`swiftsimio.reader.__SWIFTNamedColumnDataset`:
:class:`_SWIFTGroupDatasetHelper` and
:class:`_SWIFTNamedColumnDatasetHelper`, respectively. In general objects of
these types should not be created directly by users, but rather by an object of
the :class:`SWIFTGalaxy` class.
"""

from warnings import warn
from copy import deepcopy
import numpy as np
from scipy.spatial.transform import Rotation
import unyt
from swiftsimio import metadata as swiftsimio_metadata
from swiftsimio.reader import (
    SWIFTDataset,
    __SWIFTNamedColumnDataset,
    __SWIFTGroupDataset,
)
from swiftsimio.objects import cosmo_array, cosmo_factor, a
from swiftsimio.masks import SWIFTMask
from swiftgalaxy.halo_catalogues import _HaloCatalogue
from swiftgalaxy.masks import MaskCollection

from typing import Union, Any, Optional, Set


def _apply_box_wrap(coords: cosmo_array, boxsize: Optional[cosmo_array]) -> cosmo_array:
    """
    Wrap coordinates for periodic box.

    Given some coordinates, wrap the box size so that they lie within the periodic volume.

    Parameters
    ----------
    coords : :class:`~swiftsimio.objects.cosmo_array`
        The coordinates to be wrapped.

    boxsize : :class:`~swiftsimio.objects.cosmo_array` or ``None``
        The dimensions of the box to wrap (3 elements).

    Returns
    -------
    out : :class:`~swiftsimio.objects.cosmo_array`
        The coordinates wrapped to lie within the box dimensions.
    """
    return (
        (coords + boxsize / 2.0) % boxsize - boxsize / 2.0
        if boxsize is not None
        else coords
    )


def _apply_translation(coords: cosmo_array, offset: cosmo_array) -> cosmo_array:
    """
    Apply a translation to a coordinate array.

    Also warns the user of ambiguity in physical/comoving coordinates.

    Parameters
    ----------
    coords : :class:`~swiftsimio.objects.cosmo_array`
        The coordinate array to be translated.
    offset : :class:`~swiftsimio.objects.cosmo_array`
        The translation vector.

    Returns
    -------
    out : :class:`~swiftsimio.objects.cosmo_array`
        The coordinate array with the translation applied.
    """
    if hasattr(offset, "comoving") and coords.comoving:
        offset = offset.to_comoving()
    elif hasattr(offset, "comoving") and not coords.comoving:
        offset = offset.to_physical()
    elif not hasattr(offset, "comoving"):
        msg = (
            "Translation assumed to be in comoving (not physical) coordinates."
            if coords.comoving
            else "Translation assumed to be in physical (not comoving) coordinates."
        )
        warn(msg, category=RuntimeWarning)
    return coords + offset


def _apply_rotmat(coords: cosmo_array, rotation_matrix: np.ndarray) -> cosmo_array:
    """
    Apply a rotation matrix to a coordinate array.

    Applies a rotation in-place using a view through a :class:`numpy.ndarray`, then
    restores units and metadata of the :class:`~swiftsimio.objects.cosmo_array`.

    Parameters
    ----------
    coords : :class:`~swiftsimio.objects.cosmo_array`
        The coordinate array to be rotated.
    rotation_matrix : :class:`~numpy.ndarray`
        The rotation matrix (3x3).

    Returns
    -------
    out : :class:`~swiftsimio.objects.cosmo_array`
        The coordinate array with rotation applied.
    """
    return cosmo_array(
        coords.view(np.ndarray).dot(rotation_matrix),
        units=coords.units,
        cosmo_factor=coords.cosmo_factor,
        comoving=coords.comoving,
    )


def _apply_4transform(
    coords: cosmo_array, transform: np.ndarray, transform_units: unyt.unyt_quantity
) -> cosmo_array:
    """
    Apply an arbitary coordinate transformation (translation mixed with rotation) to a
    coordinate array.

    An arbitrary coordinate transformation mixing translations and rotations can be
    expressed as a 4x4 matrix. However, such a matrix has mixed units, so we need to
    assume a consistent unit for all transformations and work with bare arrays. We also
    always assume comoving coordinates.

    Parameters
    ----------
    coords : :class:`~swiftsimio.objects.cosmo_array`
        The coordinate array to be transformed.
    transform : :class:`~numpy.ndarray`
        The 4x4 transformation matrix.
    transform_units : :class:`unyt.unyt_quantity`
        The units assumed in the translation portion of the transformation matrix.

    Returns
    -------
    out : :class:`~swiftsimio.objects.cosmo_array`
        The coordinate array with transformation applied.
    """
    retval = cosmo_array(
        np.hstack(
            (
                coords.to_comoving().to_value(transform_units),
                np.ones(coords.shape[0])[:, np.newaxis],
            )
        ).dot(transform)[:, :3],
        units=transform_units,
        comoving=True,
        cosmo_factor=coords.cosmo_factor,
    )
    if coords.comoving:
        return retval.to_comoving()
    else:
        return retval.to_physical()


class _CoordinateHelper(object):
    """
    Container class for coordinates.

    Stores a dictionary of coordinate arrays and names (and aliases) for these,
    and enables accessing the arrays via :meth:`__getattr__` (dot syntax). For
    interactive use, printing a :class:`_CoordinateHelper`
    lists the available coordinate names and aliases.

    Parameters
    ----------
    coordinates : :obj:`dict` or :class:`~swiftsimio.objects.cosmo_array`
        The coordinate array(s) to be stored.

    masks : :class:`dict`
        Available coordinate names and their aliases with corresponding masks
        (or keys) into the coordinate array or dictionary for each.
    """

    def __init__(self, coordinates: Union[dict, cosmo_array], masks: dict) -> None:
        self._coordinates: Union[np.ndarray, dict] = coordinates
        self._masks: dict = masks
        return

    def __getattr__(self, attr: str) -> cosmo_array:
        """
        Get a coordinate array using attribute (dot) syntax.

        Looks up the requested attribute in the internal register of coordinate
        array names and their aliases to retrieve the array corresponding to the
        request.

        Parameters
        ----------
        attr : :obj:`str`
            The name (possibly an alias) of the coordinate array to retrieve.

        Returns
        -------
        out : :class:`~swiftsimio.objects.cosmo_array`
            The requested coordinate array.
        """
        return self._coordinates[self._masks[attr]]

    def __str__(self) -> str:
        """
        Get a string representation of the available coordinate array names.

        Returns
        -------
        out : :obj:`str`
            The string representation.
        """
        keys = ", ".join(self._masks.keys())
        return f"Available coordinates: {keys}."

    def __repr__(self) -> str:
        """
        Get a string representation of the available coordinate array names.

        Returns
        -------
        out : :obj:`str`
            The string representation.
        """
        return self.__str__()


class _SWIFTNamedColumnDatasetHelper(object):
    """
    A wrapper class to enable :class:`SWIFTGalaxy`
    functionality for a :class:`swiftsimio.reader.__SWIFTNamedColumnDataset`.

    Unlike the :class:`SWIFTGalaxy` class that inherits
    directly from :class:`~swiftsimio.reader.SWIFTDataset`, for technical
    reasons this class does *not* inherit
    :class:`swiftsimio.reader.__SWIFTNamedColumnDataset`. It does, however,
    expose the functionality of that class by maintaining an instance
    internally and forwarding any attribute lookups that it does not handle
    itself to its internal named column dataset.

    Like :class:`_SWIFTGroupDatasetHelper`, this class handles the
    transformation and masking of data from calls to :class:`SWIFTGalaxy`
    routines.

    Instances of this helper class should in general not be created separately
    since they require an instance of :class:`SWIFTGalaxy` to function and
    will be created automatically by that class.

    If any datasets contained in a named column dataset should transform like
    particle coordinates or velocities, these can be specified in the arguments
    ``transforms_like_coordinates`` and ``transforms_like_velocities`` to
    :class:`SWIFTGalaxy` as a string containing a dot, e.g. the argument
    ``transforms_like_coordinates={'coordinates',
    'extra_coordinates.an_extra_coordinate'}`` is syntactically valid.

    Parameters
    ----------
    named_column_dataset : :class:`swiftsimio.reader.__SWIFTNamedColumnDataset`
        The named column dataset to be wrapped.

    particle_dataset_helper : :class:`_SWIFTGroupDatasetHelper`
        Used to store a reference to the parent
        :class:`_SWIFTGroupDatasetHelper` object.

    See Also
    --------
    :class:`SWIFTGalaxy`
    :class:`_SWIFTGroupDatasetHelper`
    """

    def __init__(
        self,
        named_column_dataset: "__SWIFTNamedColumnDataset",
        particle_dataset_helper: "_SWIFTGroupDatasetHelper",
    ) -> None:
        self._named_column_dataset: __SWIFTNamedColumnDataset = named_column_dataset
        self._particle_dataset_helper: "_SWIFTGroupDatasetHelper" = (
            particle_dataset_helper
        )
        self._initialised = True
        return

    def __str__(self) -> str:
        """
        Get a string representation of the object.

        Delegated to the underlying :class:`~swiftsimio.reader.__SWIFTNamedColumnDataset`.

        Returns
        -------
        out : :obj:`str`
            The string representation.
        """
        return str(self._named_column_dataset)

    def __repr__(self) -> str:
        """
        Get a string representation of the object.

        Delegated to the underlying :class:`~swiftsimio.reader.__SWIFTNamedColumnDataset`.

        Returns
        -------
        out : :obj:`str`
            The string representation.
        """
        return self.__str__()

    def __getattribute__(self, attr: str) -> Any:
        """
        Get an attribute of the object.

        Redirects requests for datasets to the underlying
        :class:`~swiftsimio.reader.__SWIFTNamedColumnDataset` and then handles
        any transformations and masking when new data are read.

        Parameters
        ----------
        attr : :obj:`str`
            The name of the attribute.

        Returns
        -------
        out : :obj:`object`
            The value of the attribute.
        """
        named_column_dataset = object.__getattribute__(self, "_named_column_dataset")
        particle_dataset_helper = object.__getattribute__(
            self, "_particle_dataset_helper"
        )
        if attr in named_column_dataset.named_columns:
            # we're dealing with one of the named columns
            if getattr(named_column_dataset, f"_{attr}") is None:
                # going to read from file: apply masks, transforms
                if (
                    hasattr(particle_dataset_helper._swiftgalaxy, "_initialised")
                    and particle_dataset_helper._swiftgalaxy._warn_on_read
                ):
                    warn(
                        f"Reading {particle_dataset_helper.group_name}."
                        f"{named_column_dataset.name}.{attr} from snapshot file, "
                        "this may be unintended (should it be preloaded if using "
                        "SWIFTGalaxies to iterate over objects of interest?)",
                        RuntimeWarning,
                    )
                data = getattr(named_column_dataset, attr)  # raw data loaded
                data = particle_dataset_helper._apply_data_mask(data)
                data = particle_dataset_helper._apply_transforms(
                    data, f"{named_column_dataset.name}.{attr}"
                )
                setattr(named_column_dataset, f"_{attr}", data)
            else:
                # just return the data
                pass
        try:
            # beware collisions with SWIFTGroupDataset namespace
            return object.__getattribute__(self, attr)
        except AttributeError:
            # exposes everything else in __dict__
            return getattr(named_column_dataset, attr)

    def __setattr__(self, attr: str, value: Any) -> None:
        """
        Set an attribute of the object.

        Redirects setting attributes to the underlying
        :class:`~swiftsimio.reader.__SWIFTNamedColumnDataset` objects when relevant.

        Parameters
        ----------
        attr : :obj:`str`
            Name of the attribute.
        value : :obj:`object`
            The value to assign to the attribute.
        """
        # pass particle data through to actual SWIFTNamedColumnDataset
        if not hasattr(self, "_initialised"):
            # guard during initialisation
            object.__setattr__(self, attr, value)
            return
        column_names = self._named_column_dataset.named_columns
        if (attr in column_names) or (
            (attr.startswith("_")) and (attr[1:] in column_names)
        ):
            setattr(self._named_column_dataset, attr, value)
            return
        else:
            object.__setattr__(self, attr, value)
            return

    def __getitem__(self, mask: slice) -> "_SWIFTNamedColumnDatasetHelper":
        """
        Apply a mask to the :class:`~swiftgalaxy.reader._SWIFTNamedColumnDatasetHelper`
        with square-bracket notation.

        To ensure internal consistency this requires producing a full ("deep") copy of
        the parent :class:`~swiftgalaxy.reader.SWIFTGalaxy` object and all of its contents
        (but data is masked at copy time to avoid unnecessary memory overhead).

        Parameters
        ----------
        mask : :obj:`slice`
            The mask to apply to the named column data arrays (and all other data arrays
            for particles of the same type).
        """
        return self._data_copy(mask=mask)

    def __copy__(self) -> "_SWIFTNamedColumnDatasetHelper":
        """
        Create a copy of the :class:`~swiftgalaxy.reader._SWIFTNamedColumnDatasetHelper`
        without copying data (a "shallow" copy).

        Returns
        -------
        out : :class:`~swiftgalaxy.reader._SWIFTNamedColumnDatasetHelper`
            The copy of the :class:`~swiftgalaxy.reader._SWIFTNamedColumnDatasetHelper`
            object.
        """
        return getattr(self._particle_dataset_helper.__copy__(), self.name)

    def __deepcopy__(
        self, memo: Optional[dict] = None
    ) -> "_SWIFTNamedColumnDatasetHelper":
        """
        Create a copy of the :class:`~swiftgalaxy.reader._SWIFTNamedColumnDatasetHelper`
        including copying data (a "deep" copy).

        Parameters
        ----------
        memo : :obj:`dict` (optional), default: ``None``
            For the copy operation to keep a record of already copied objects.

        Returns
        -------
        out : :class:`~swiftgalaxy.reader._SWIFTNamedColumnDatasetHelper`
            The copy of the :class:`~swiftgalaxy.reader.SWIFTGalaxy` object.
        """

        return self._data_copy()

    def _data_copy(
        self, mask: Optional[slice] = None
    ) -> "_SWIFTNamedColumnDatasetHelper":
        """
        Create a copy of the :class:`~swiftgalaxy.reader._SWIFTNamedColumnDatasetHelper`
        including copying data (a "deep" copy).

        Parameters
        ----------
        mask : :obj:`slice` (optional), default: ``None``
            Copy only the subset of the data corresponding to the ``mask``.

        Returns
        -------
        out : :class:`~swiftgalaxy.reader._SWIFTNamedColumnDatasetHelper`
            A (possibly masked) copy of the
            :class:`~swiftgalaxy.reader._SWIFTNamedColumnDatasetHelper` object.
        """
        return getattr(self._particle_dataset_helper._data_copy(mask=mask), self.name)


class _SWIFTGroupDatasetHelper(object):
    """
    A wrapper class to enable :class:`SWIFTGalaxy`
    functionality for a :class:`swiftsimio.reader.__SWIFTGroupDatasets`.

    Unlike the :class:`SWIFTGalaxy` class that inherits
    directly from :class:`~swiftsimio.reader.SWIFTDataset`, for technical
    reasons this class does *not* inherit
    :class:`swiftsimio.reader.__SWIFTGroupDatasets`. It does, however,
    expose the functionality of that class by maintaining an instance
    internally and forwarding any attribute lookups that it does not handle
    itself to its internal particle dataset.

    In addition to handling the transformation and masking of data from calls
    to :class:`SWIFTGalaxy` routines, this class provides
    particle coordinates and velocities in cartesian, spherical and cylindrical
    coordinates through the properties:

    + :attr:`cartesian_coordinates`
    + :attr:`cartesian_velocities`
    + :attr:`spherical_coordinates`
    + :attr:`spherical_velocities`
    + :attr:`cylindrical_coordinates`
    + :attr:`cylindrical_velocities`

    These are evaluated lazily and automatically re-calculated if necessary,
    such as after a coordinate rotation.

    Instances of this helper class should in general not be created separately
    since they require an instance of :class:`SWIFTGalaxy` to function and
    will be created automatically by that class.

    Parameters
    ----------
    particle_dataset : :class:`swiftsimio.reader.__SWIFTGroupDatasets`
        The particle dataset to be wrapped.

    swiftgalaxy : :class:`SWIFTGalaxy`
        Used to store a reference to the parent :class:`SWIFTGalaxy`.

    See Also
    --------
    :class:`SWIFTGalaxy`
    :class:`_CoordinateHelper`

    Examples
    --------

    The cartesian, spherical and cylindrical coordinates of gas particles can
    be accessed, for example, by (``mygalaxy`` is a :class:`SWIFTGalaxy`):

    ::

        mygalaxy.gas.cartesian_coordinates.x
        mygalaxy.gas.cartesian_coordinates.y
        mygalaxy.gas.cartesian_coordinates.z
        mygalaxy.gas.cartesian_velocities.x
        mygalaxy.gas.cartesian_velocities.y
        mygalaxy.gas.cartesian_velocities.z
        mygalaxy.gas.spherical_coordinates.r
        mygalaxy.gas.spherical_coordinates.theta
        mygalaxy.gas.spherical_coordinates.phi
        mygalaxy.gas.spherical_velocities.r
        mygalaxy.gas.spherical_velocities.theta
        mygalaxy.gas.spherical_velocities.phi
        mygalaxy.gas.cylindrical_coordinates.rho
        mygalaxy.gas.cylindrical_coordinates.phi
        mygalaxy.gas.cylindrical_coordinates.z
        mygalaxy.gas.cylindrical_velocities.rho
        mygalaxy.gas.cylindrical_velocities.phi
        mygalaxy.gas.cylindrical_velocities.z
    """

    def __init__(
        self, particle_dataset: "__SWIFTGroupDataset", swiftgalaxy: "SWIFTGalaxy"
    ) -> None:
        self._particle_dataset: __SWIFTGroupDataset = particle_dataset
        self._swiftgalaxy: "SWIFTGalaxy" = swiftgalaxy
        self._named_column_dataset_helpers: dict[
            str, _SWIFTNamedColumnDatasetHelper
        ] = dict()
        particle_metadata = getattr(self.metadata, f"{self.group_name}_properties")
        named_columns_names = [
            fn
            for (fn, fp) in zip(
                particle_metadata.field_names, particle_metadata.field_paths
            )
            if particle_metadata.named_columns[fp] is not None
        ]
        for named_columns_name in named_columns_names:
            # This is the named_columns instance to wrap:
            named_columns = getattr(self._particle_dataset, named_columns_name)
            # We'll make a custom type to present a nice name to the user.
            particle_nice_name = swiftsimio_metadata.particle_types.particle_name_class[
                self.group
            ]
            nice_name = (
                f"{particle_nice_name}"
                f"{named_columns.field_path.split('/')[-1]}ColumnsHelper"
            )
            TypeNamedColumnDatasetHelper = type(
                nice_name, (_SWIFTNamedColumnDatasetHelper, object), dict()
            )
            self._named_column_dataset_helpers[named_columns_name] = (
                TypeNamedColumnDatasetHelper(named_columns, self)
            )
        self._spherical_coordinates: Optional[dict] = None
        self._cylindrical_coordinates: Optional[dict] = None
        self._spherical_velocities: Optional[dict] = None
        self._cylindrical_velocities: Optional[dict] = None
        self._initialised = True
        return

    def __str__(self) -> str:
        """
        Get a string representation of the object.

        Delegated to the underlying :class:`~swiftsimio.reader.__SWIFTGroupDataset`.

        Returns
        -------
        out : :obj:`str`
            The string representation.
        """
        return str(self._particle_dataset)

    def __repr__(self) -> str:
        """
        Get a string representation of the object.

        Delegated to the underlying :class:`~swiftsimio.reader.__SWIFTGroupDataset`.

        Returns
        -------
        out : :obj:`str`
            The string representation.
        """
        return self.__str__()

    def __getattribute__(self, attr: str) -> Any:
        """
        Get an attribute of the object.

        Redirects requests for datasets to the underlying
        :class:`~swiftsimio.reader.__SWIFTGroupDataset` and then handles
        any transformations and masking when new data are read.

        Parameters
        ----------
        attr : :obj:`str`
            The name of the attribute.

        Returns
        -------
        out : :obj:`object`
            The value of the attribute.
        """
        particle_name = object.__getattribute__(self, "_particle_dataset").group_name
        metadata = object.__getattribute__(self, "_particle_dataset").metadata
        particle_metadata = getattr(metadata, f"{particle_name}_properties")
        particle_dataset = object.__getattribute__(self, "_particle_dataset")
        if attr in particle_metadata.field_names:
            # check if we're dealing with a named columns field
            if object.__getattribute__(self, "_is_namedcolumns")(attr):
                return object.__getattribute__(self, "_named_column_dataset_helpers")[
                    attr
                ]
            # otherwise we're dealing with a particle data table
            if getattr(particle_dataset, f"_{attr}") is None:
                if (
                    hasattr(
                        object.__getattribute__(self, "_swiftgalaxy"), "_initialised"
                    )
                    and object.__getattribute__(self, "_swiftgalaxy")._warn_on_read
                ):
                    warn(
                        f"Reading {particle_name}.{attr} from snapshot file, this may be "
                        "unintended (should it be preloaded if using SWIFTGalaxies to "
                        "iterate over objects of interest?)",
                        RuntimeWarning,
                    )
                # going to read from file: apply masks, transforms
                data = getattr(particle_dataset, attr)  # raw data loaded
                data = object.__getattribute__(self, "_apply_data_mask")(data)
                data = object.__getattribute__(self, "_apply_transforms")(data, attr)
                setattr(particle_dataset, f"_{attr}", data)
            else:
                # just return the data
                pass
        try:
            # beware collisions with SWIFTDataset namespace
            return object.__getattribute__(self, attr)
        except AttributeError:
            # exposes everything else in __dict__
            return getattr(particle_dataset, attr)

    def __setattr__(self, attr: str, value: Any) -> None:
        """
        Set an attribute of the object.

        Redirects setting attributes to the underlying
        :class:`~swiftsimio.reader.__SWIFTGroupDataset` objects when relevant.

        Parameters
        ----------
        attr : :obj:`str`
            Name of the attribute.
        value : :obj:`object`
            The value to assign to the attribute.
        """
        # pass particle data through to actual SWIFTDataset
        if not hasattr(self, "_initialised"):
            # guard during initialisation
            object.__setattr__(self, attr, value)
            return
        field_names = getattr(
            self._particle_dataset.metadata,
            f"{self._particle_dataset.group_name}_properties",
        ).field_names
        if (attr in field_names) and self._is_namedcolumns(attr):
            self._named_column_dataset_helpers[attr] = value
            return
        elif (attr in field_names) or (
            (attr.startswith("_")) and (attr[1:] in field_names)
        ):
            setattr(self._particle_dataset, attr, value)
            return
        else:
            object.__setattr__(self, attr, value)
            return

    def __getitem__(self, mask: slice) -> "_SWIFTGroupDatasetHelper":
        """
        Apply a mask to the :class:`~swiftgalaxy.reader._SWIFTGroupDatasetHelper`
        with square-bracket notation.

        To ensure internal consistency this requires producing a full ("deep") copy of
        the parent :class:`~swiftgalaxy.reader.SWIFTGalaxy` object and all of its contents
        (but data is masked at copy time to avoid unnecessary memory overhead).

        Parameters
        ----------
        mask : :obj:`slice`
            The mask to apply to the particle data arrays.
        """
        return self._data_copy(mask=mask)

    def __copy__(self) -> "_SWIFTGroupDatasetHelper":
        """
        Create a copy of the :class:`~swiftgalaxy.reader._SWIFTGroupDatasetHelper`
        without copying data (a "shallow" copy).

        Returns
        -------
        out : :class:`~swiftgalaxy.reader._SWIFTGroupDatasetHelper`
            The copy of the :class:`~swiftgalaxy.reader._SWIFTGroupDatasetHelper` object.
        """
        return getattr(self._swiftgalaxy.__copy__(), self.particle_name)

    def __deepcopy__(self, memo: Optional[dict] = None) -> "_SWIFTGroupDatasetHelper":
        """
        Create a copy of the :class:`~swiftgalaxy.reader._SWIFTGroupDatasetHelper`
        including copying data (a "deep" copy).

        Parameters
        ----------
        memo : :obj:`dict` (optional), default: ``None``
            For the copy operation to keep a record of already copied objects.

        Returns
        -------
        out : :class:`~swiftgalaxy.reader._SWIFTGroupDatasetHelper`
            The copy of the :class:`~swiftgalaxy.reader._SWIFTGroupDatasetHelper` object.
        """
        return self._data_copy()

    def _data_copy(self, mask: Optional[slice] = None) -> "_SWIFTGroupDatasetHelper":
        """
        Create a copy of the :class:`~swiftgalaxy.reader._SWIFTGroupDatasetHelper`
        including copying data (a "deep" copy).

        Parameters
        ----------
        mask : :obj:`slice` (optional), default: ``None``
            Copy only the subset of the data corresponding to the ``mask``.

        Returns
        -------
        out : :class:`~swiftgalaxy.reader._SWIFTGroupDatasetHelper`
            A (possibly masked) copy of the
            :class:`~swiftgalaxy.reader._SWIFTGroupDatasetHelper` object.
        """
        mask_collection = MaskCollection(
            **{
                k: None if k != self.group_name else mask
                for k in self.metadata.present_group_names
            }
        )
        return getattr(
            self._swiftgalaxy._data_copy(mask_collection=mask_collection),
            self.group_name,
        )

    def _is_namedcolumns(self, field_name: str) -> bool:
        """
        Checks a string against the metadata to determine if it describes a named column.

        Parameters
        ----------
        field_name : :obj:`str`
            The name of the field to check against the list of named column datasets.

        Returns
        -------
        out : :obj:`bool`
            ``True`` if ``field_name`` describes a named column, else ``False``.
        """
        particle_name = self._particle_dataset.group_name
        particle_metadata = getattr(
            self._particle_dataset.metadata, f"{particle_name}_properties"
        )
        field_path = dict(
            zip(particle_metadata.field_names, particle_metadata.field_paths)
        )[field_name]
        return particle_metadata.named_columns[field_path] is not None

    def _apply_data_mask(self, data: cosmo_array) -> cosmo_array:
        """
        Used internally to apply existing masks on reading new data.

        Parameters
        ----------
        data : :class:`~swiftsimio.objects.cosmo_array`
            The data to mask.

        Returns
        -------
        out : :class:`~swiftsimio.objects.cosmo_array`
            The data with any masks applied.
        """
        if self._swiftgalaxy._extra_mask is not None:
            mask = getattr(
                self._swiftgalaxy._extra_mask, self._particle_dataset.group_name
            )
            if mask is not None:
                return data[mask]
        return data

    def _mask_dataset(self, mask: slice) -> None:
        """
        Apply a mask to this data set.

        Intended for internal use.

        Parameters
        ----------
        mask : :obj:`slice`
            The mask to apply to all data arrays managed by this
            :class:`~swiftgalaxy.reader._SWIFTGroupDatasetHelper`.
        """
        # Users are cautioned against calling this function directly!
        # Use SWIFTGalaxy.mask_particles instead.
        particle_name = self._particle_dataset.group_name
        particle_metadata = getattr(
            self._particle_dataset.metadata, f"{particle_name}_properties"
        )
        for field_name in particle_metadata.field_names:
            if self._is_namedcolumns(field_name):
                for named_column in getattr(self, field_name).named_columns:
                    if (
                        getattr(getattr(self, field_name), f"_{named_column}")
                        is not None
                    ):
                        setattr(
                            getattr(self, field_name),
                            f"_{named_column}",
                            getattr(getattr(self, field_name), f"_{named_column}")[
                                mask
                            ],
                        )
            elif getattr(self, f"_{field_name}") is not None:
                setattr(self, f"_{field_name}", getattr(self, f"_{field_name}")[mask])
        self._mask_derived_coordinates(mask)
        if getattr(self._swiftgalaxy._extra_mask, particle_name) is None:
            setattr(self._swiftgalaxy._extra_mask, particle_name, mask)
        else:
            if self._swiftgalaxy._spatial_mask is None:
                # get a count of particles in the box
                num_part = self._particle_dataset.metadata.num_part[
                    particle_metadata.particle_type
                ]
            else:
                # get a count of particles in the spatial mask region
                num_part = np.sum(
                    self._swiftgalaxy._spatial_mask.get_masked_counts_offsets()[0][
                        particle_name
                    ]
                )
            old_mask = getattr(self._swiftgalaxy._extra_mask, particle_name)
            # need to convert to an integer mask to combine
            # (boolean is insufficient in case of re-ordering masks)
            setattr(
                self._swiftgalaxy._extra_mask,
                particle_name,
                np.arange(num_part, dtype=int)[old_mask][mask],
            )
        return

    def _apply_transforms(self, data: cosmo_array, dataset_name: str) -> cosmo_array:
        """
        Used internally to apply existing coordinate transforms on reading new data.

        Checks whether the input dataset_name is in the list of datasets that need
        to have coordinate transformation (either coordinate-like or velocity-like)
        and applies the transformations as needed.

        Parameters
        ----------
        data : :class:`~swiftsimio.objects.cosmo_array`
            The data to (potentially) transform.
        dataset_name : :obj:`str`
            The name of the dataset contained in ``data``.

        Returns
        -------
        out : :class:`~swiftsimio.objects.cosmo_array`
            The data with any required transformations applied.
        """
        if dataset_name in self._swiftgalaxy.transforms_like_coordinates:
            transform_units = self._swiftgalaxy.metadata.units.length
            transform = self._swiftgalaxy._coordinate_like_transform
        elif dataset_name in self._swiftgalaxy.transforms_like_velocities:
            transform_units = (
                self._swiftgalaxy.metadata.units.length
                / self._swiftgalaxy.metadata.units.time
            )
            transform = self._swiftgalaxy._velocity_like_transform
        else:
            transform = None
        if transform is not None:
            data = _apply_4transform(data, transform, transform_units)
        boxsize = getattr(self._particle_dataset.metadata, "boxsize", None)
        if dataset_name in self._swiftgalaxy.transforms_like_coordinates:
            data = _apply_box_wrap(data, boxsize)
        return data

    @property
    def cartesian_coordinates(self) -> _CoordinateHelper:
        """
        Utility to access the cartesian coordinates of particles.

        Returns a wrapper around the coordinate array which can be accessed using
        attribute syntax. Cartesian coordinates can be accessed separately:

        + ``cartesian_coordinates.x``
        + ``cartesian_coordinates.y``
        + ``cartesian_coordinates.z``

        or as a 2D array:

        + ``cartesian_coordinates.xyz``

        A reference to the coordinates array is obtained each time, so cartesian
        coordinates are automatically updated if the coordinates array is modified (e.g.
        following a rotation or other transformation).

        By default the coorinate array is assumed to be called ``coordinates``,
        but this can be overridden with the ``coordinates_dataset_name``
        argument to :class:`SWIFTGalaxy`.

        Returns
        -------
        coordinate_helper : :class:`_CoordinateHelper`
            Container providing particle cartesian coordinates as attributes.
        """
        return _CoordinateHelper(
            getattr(self, self._swiftgalaxy.coordinates_dataset_name),
            dict(x=np.s_[:, 0], y=np.s_[:, 1], z=np.s_[:, 2], xyz=np.s_[...]),
        )

    @property
    def cartesian_velocities(self) -> _CoordinateHelper:
        """
        Utility to access the cartesian components of particle velocities.

        Returns a wrapper around the velocities array which can be accessed using
        attribute syntax. Cartesian coordinates can be accessed separately:

        + ``cartesian_velocities.x``
        + ``cartesian_velocities.y``
        + ``cartesian_velocities.z``

        or as a 2D array:

        + ``cartesian_velocities.xyz``

        A reference to the velocities array is obtained each time, so cartesian
        velocities are automatically updated if the velocities array is modified (e.g.
        following a rotation or other transformation).

        By default the array of velocities is assumed to be called
        ``velocities``, but this can be overridden with the
        ``velocities_dataset_name`` argument to :class:`SWIFTGalaxy`.

        Returns
        -------
        coordinate_helper : :class:`_CoordinateHelper`
            Container providing particle cartesian velocities as attributes.
        """
        return _CoordinateHelper(
            getattr(self, self._swiftgalaxy.velocities_dataset_name),
            dict(x=np.s_[:, 0], y=np.s_[:, 1], z=np.s_[:, 2], xyz=np.s_[...]),
        )

    @property
    def spherical_coordinates(self) -> _CoordinateHelper:
        """
        Utility to access the spherical coordinates of particles.

        The spherical coordinates of particles are calculated the first time
        this attribute is accessed. If a coordinate transformation (e.g. a
        rotation) or other operation is applied to the :class:`SWIFTGalaxy`
        that would invalidate the derived spherical coordinates, they are
        erased and will be recalculated at the next access of this attribute.
        The coordinates could be transformed when they change instead, but in
        general this requires transforming back through cartesian coordinates,
        so the more efficient "lazy" approach of recalculating on demand is
        used instead.

        The "physics" notation convention, where
        :math:`-\\frac{\\pi}{2} \\leq \\theta \\leq \\frac{\\pi}{2}` is the
        polar angle and :math:`0 < \\phi \\leq 2\\pi` is the azimuthal angle,
        is assumed.

        Several attribute names are supported for each coordinate. They can be
        accessed with the aliases:

        + ``spherical_coordinates.r``:
            + ``spherical_coordinates.radius``
        + ``spherical_coordinates.theta``:
            + ``spherical_coordinates.lat``
            + ``spherical_coordinates.latitude``
            + ``spherical_coordinates.pol``
            + ``spherical_coordinates.polar``
        + ``spherical_coordinates.phi``:
            + ``spherical_coordinates.lon``
            + ``spherical_coordinates.longitude``
            + ``spherical_coordinates.az``
            + ``spherical_coordinates.azimuth``

        By default the coorinate array is assumed to be called ``coordinates``,
        but this can be overridden with the ``coordinates_dataset_name``
        argument to :class:`SWIFTGalaxy`.

        Returns
        -------
        coordinate_helper : :class:`_CoordinateHelper`
            Container providing particle spherical coordinates as attributes.
        """
        if self._spherical_coordinates is None:
            r = np.sqrt(np.sum(np.power(self.cartesian_coordinates.xyz, 2), axis=1))
            theta = cosmo_array(
                np.where(r == 0, 0, np.arcsin(self.cartesian_coordinates.z / r)),
                units=unyt.rad,
                comoving=r.comoving,
                cosmo_factor=cosmo_factor(
                    a**0, scale_factor=r.cosmo_factor.scale_factor
                ),
            )
            if self.cylindrical_coordinates is not None:
                phi = self.cylindrical_coordinates.phi
            else:
                phi = cosmo_array(
                    np.arctan2(
                        self.cartesian_coordinates.y, self.cartesian_coordinates.x
                    ),
                    units=unyt.rad,
                    comoving=r.comoving,
                    cosmo_factor=cosmo_factor(
                        a**0, scale_factor=r.cosmo_factor.scale_factor
                    ),
                )
                phi[phi < 0] = phi[phi < 0] + 2 * np.pi * unyt.rad
            self._spherical_coordinates = dict(_r=r, _theta=theta, _phi=phi)
        return _CoordinateHelper(
            self._spherical_coordinates,
            dict(
                r="_r",
                radius="_r",
                lon="_phi",
                longitude="_phi",
                az="_phi",
                azimuth="_phi",
                phi="_phi",
                lat="_theta",
                latitude="_theta",
                pol="_theta",
                polar="_theta",
                theta="_theta",
            ),
        )

    @property
    def spherical_velocities(self) -> _CoordinateHelper:
        """
        Utility to access the velocities of particles in spherical coordinates.

        The particle velocities in spherical coordinates are calculated the
        first time this attribute is accessed. If a coordinate transformation
        (e.g. a rotation) or other operation is applied to the
        :class:`SWIFTGalaxy` that would invalidate the derived spherical
        velocities, they are erased and will be recalculated at the next access
        of this attribute. The velocities could be transformed when they change
        instead, but in general this requires transforming back through
        cartesian coordinates, so the more efficient "lazy" approach of
        recalculating on demand is used instead.

        The "physics" notation convention, where
        :math:`-\\frac{\\pi}{2} \\leq \\theta \\leq \\frac{\\pi}{2}` is the
        polar angle and :math:`0 < \\phi \\leq 2\\pi` is the azimuthal angle,
        is assumed.

        Several attribute names are supported for each velocity component. They
        can be accessed with the aliases:

        + ``spherical_velocities.r``:
            + ``spherical_velocities.radius``
        + ``spherical_velocities.theta``:
            + ``spherical_velocities.lat``
            + ``spherical_velocities.latitude``
            + ``spherical_velocities.pol``
            + ``spherical_velocities.polar``
        + ``spherical_velocities.phi``:
            + ``spherical_velocities.lon``
            + ``spherical_velocities.longitude``
            + ``spherical_velocities.az``
            + ``spherical_velocities.azimuth``

        By default the array of velocities is assumed to be called
        ``velocities``, but this can be overridden with the
        ``velocities_dataset_name`` argument to :class:`SWIFTGalaxy`.

        Returns
        -------
        coordinate_helper : :class:`_CoordinateHelper`
            Container providing particle velocities in spherical coordinates as
            attributes.
        """
        if self._spherical_coordinates is None:
            self.spherical_coordinates
        if self._spherical_velocities is None:
            _sin_t = np.sin(self.spherical_coordinates.theta)
            _cos_t = np.cos(self.spherical_coordinates.theta)
            _sin_p = np.sin(self.spherical_coordinates.phi)
            _cos_p = np.cos(self.spherical_coordinates.phi)
            v_r = (
                _cos_t * _cos_p * self.cartesian_velocities.x
                + _cos_t * _sin_p * self.cartesian_velocities.y
                + _sin_t * self.cartesian_velocities.z
            )
            v_t = (
                _sin_t * _cos_p * self.cartesian_velocities.x
                + _sin_t * _sin_p * self.cartesian_velocities.y
                - _cos_t * self.cartesian_velocities.z
            )
            v_p = (
                -_sin_p * self.cartesian_velocities.x
                + _cos_p * self.cartesian_velocities.y
            )
            self._spherical_velocities = dict(_v_r=v_r, _v_t=v_t, _v_p=v_p)
        return _CoordinateHelper(
            self._spherical_velocities,
            dict(
                r="_v_r",
                radius="_v_r",
                lon="_v_p",
                longitude="_v_p",
                az="_v_p",
                azimuth="_v_p",
                phi="_v_p",
                lat="_v_t",
                latitude="_v_t",
                pol="_v_t",
                polar="_v_t",
                theta="_v_t",
            ),
        )

    @property
    def cylindrical_coordinates(self) -> _CoordinateHelper:
        """
        Utility to access the cylindrical coordinates of particles.

        The cylindrical coordinates of particles are calculated the first time
        this attribute is accessed. If a coordinate transformation (e.g. a
        rotation) or other operation is applied to the :class:`SWIFTGalaxy`
        that would invalidate the derived cylindrical coordinates, they are
        erased and will be recalculated at the next access of this attribute.
        The coordinates could be transformed when they change instead, but in
        general this requires transforming back through cartesian coordinates,
        so the more efficient "lazy" approach of recalculating on demand is
        used instead.

        The coordinate components are named :math:`(\\rho, \\phi, z)` by
        default, and assume a convention where :math:`0 < \\phi \\leq 2\\pi`.

        Several attribute names are supported for each coordinate. They can be
        accessed with the aliases:

        + ``cylindrical_coordinates.rho``:
            + ``cylindrical_coordinates.R``
            + ``cylindrical_coordinates.radius``
        + ``cylindrical_coordinates.phi``:
            + ``cylindrical_coordinates.lon``
            + ``cylindrical_coordinates.longitude``
            + ``cylindrical_coordinates.az``
            + ``cylindrical_coordinates.azimuth``
        + ``cylindrical_coordinates.z``

        By default the coorinate array is assumed to be called ``coordinates``,
        but this can be overridden with the ``coordinates_dataset_name``
        argument to :class:`SWIFTGalaxy`.

        Returns
        -------
        coordinate_helper : :class:`_CoordinateHelper`
            Container providing particle cylindrical coordinates as attributes.
        """
        if self._cylindrical_coordinates is None:
            rho = np.sqrt(
                np.sum(np.power(self.cartesian_coordinates.xyz[:, :2], 2), axis=1)
            )
            if self._spherical_coordinates is not None:
                phi = self.spherical_coordinates.phi
            else:
                # np.where returns ndarray
                phi = np.arctan2(
                    self.cartesian_coordinates.y, self.cartesian_coordinates.x
                ).view(np.ndarray)
                phi = np.where(phi < 0, phi + 2 * np.pi, phi)
                phi = cosmo_array(
                    phi,
                    units=unyt.rad,
                    comoving=rho.comoving,
                    cosmo_factor=cosmo_factor(
                        a**0, scale_factor=rho.cosmo_factor.scale_factor
                    ),
                )
            z = self.cartesian_coordinates.z
            self._cylindrical_coordinates = dict(_rho=rho, _phi=phi, _z=z)
        return _CoordinateHelper(
            self._cylindrical_coordinates,
            dict(
                R="_rho",
                rho="_rho",
                radius="_rho",
                lon="_phi",
                longitude="_phi",
                az="_phi",
                azimuth="_phi",
                phi="_phi",
                z="_z",
                height="_z",
            ),
        )

    @property
    def cylindrical_velocities(self) -> _CoordinateHelper:
        """
        Utility to access the velocities of particles in cylindrical
        coordinates.

        The particle velocities in cylindrical coordinates are calculated the
        first time this attribute is accessed. If a coordinate transformation
        (e.g. a rotation) or other operation is applied to the
        :class:`SWIFTGalaxy` that would invalidate the derived cylindrical
        velocities, they are erased and will be recalculated at the next access
        of this attribute. The velocities could be transformed when they change
        instead, but in general this requires transforming back through
        cartesian coordinates, so the more efficient "lazy" approach of
        recalculating on demand is used instead.

        The "physics" notation convention, where
        :math:`-\\frac{\\pi}{2} \\leq \\theta \\leq \\frac{\\pi}{2}` is the
        polar angle and :math:`0 < \\phi \\leq 2\\pi` is the azimuthal angle,
        is assumed.

        The coordinate components are named :math:`(\\rho, \\phi, z)` by
        default, and assume a convention where :math:`0 < \\phi \\leq 2\\pi`.

        Several attribute names are supported for each velocity component. They
        can be accessed with the aliases:

        + ``cylindrical_velocities.rho``:
            + ``cylindrical_velocities.R``
            + ``cylindrical_velocities.radius``
        + ``cylindrical_coordinates.phi``:
            + ``cylindrical_velocities.lon``
            + ``cylindrical_velocities.longitude``
            + ``cylindrical_velocities.az``
            + ``cylindrical_velocities.azimuth``
        + ``cylindrical_velocities.z``

        By default the array of velocities is assumed to be called
        ``velocities``, but this can be overridden with the
        ``velocities_dataset_name`` argument to :class:`SWIFTGalaxy`.

        Returns
        -------
        coordinate_helper : :class:`_CoordinateHelper`
            Container providing particle velocities in cylindrical coordinates
            as attributes.
        """
        if self._cylindrical_coordinates is None:
            self.cylindrical_coordinates
        if self._cylindrical_velocities is None:
            _sin_p = np.sin(self.cylindrical_coordinates.phi)
            _cos_p = np.cos(self.cylindrical_coordinates.phi)
            v_rho = (
                _cos_p * self.cartesian_velocities.x
                + _sin_p * self.cartesian_velocities.y
            )
            if self._spherical_velocities is not None:
                v_phi = self.spherical_velocities.phi
            else:
                v_phi = (
                    -_sin_p * self.cartesian_velocities.x
                    + _cos_p * self.cartesian_velocities.y
                )
            v_z = self.cartesian_velocities.z
            self._cylindrical_velocities = dict(_v_rho=v_rho, _v_phi=v_phi, _v_z=v_z)
        return _CoordinateHelper(
            self._cylindrical_velocities,
            dict(
                R="_v_rho",
                rho="_v_rho",
                radius="_v_rho",
                lon="_v_phi",
                longitude="_v_phi",
                az="_v_phi",
                azimuth="_v_phi",
                phi="_v_phi",
                z="_v_z",
                height="_v_z",
            ),
        )

    def _mask_derived_coordinates(self, mask: slice) -> None:
        """
        Apply a mask to the internally maintained derived coordinates.

        Intended for internal use. If the user applies a mask we don't need to
        re-evaluate the derived coordinates but just apply the mask.

        Parameters
        ----------
        mask : :obj:`slice`
            Mask to apply to the derived coordinates maintained by this
            :class:`~swiftgalaxy.reader._SWIFTGroupDatasetHelper`.
        """
        if self._spherical_coordinates is not None:
            for coord in ("r", "theta", "phi"):
                self._spherical_coordinates[f"_{coord}"] = self._spherical_coordinates[
                    f"_{coord}"
                ][mask]
        if self._spherical_velocities is not None:
            for coord in ("v_r", "v_t", "v_p"):
                self._spherical_velocities[f"_{coord}"] = self._spherical_velocities[
                    f"_{coord}"
                ][mask]
        if self._cylindrical_coordinates is not None:
            for coord in ("rho", "phi", "z"):
                self._cylindrical_coordinates[f"_{coord}"] = (
                    self._cylindrical_coordinates[f"_{coord}"][mask]
                )
        if self._cylindrical_velocities is not None:
            for coord in ("v_rho", "v_phi", "v_z"):
                self._cylindrical_velocities[f"_{coord}"] = (
                    self._cylindrical_velocities[f"_{coord}"][mask]
                )
        return

    def _void_derived_coordinates(self) -> None:
        """
        Reset internal references to spherical/cylindrical coordinates to None (e.g.
        because they are no longer valid).
        """
        self._spherical_coordinates = None
        self._cylindrical_coordinates = None
        self._spherical_velocities = None
        self._cylindrical_velocities = None
        return


class SWIFTGalaxy(SWIFTDataset):
    """
    A representation of a simulated galaxy.

    A :class:`SWIFTGalaxy` represents a galaxy from a
    simulation, including both its particles and integrated properties. A halo
    finder catalogue is required to define which particles belong to the galaxy
    and to provide integrated properties. The implementation is an extension of
    the :class:`~swiftsimio.reader.SWIFTDataset` class, so all the
    functionality of such a dataset is also available for a
    :class:`SWIFTGalaxy`. The :class:`swiftsimio.reader.__SWIFTGroupDatasets`
    objects familiar to :mod:`swiftsimio` users (e.g. a ``GasDataset``) are
    wrapped by a :class:`~swiftgalaxy.reader._SWIFTGroupDatasetHelper` class that exposes
    their usual functionality and extends it with new features.
    :class:`swiftsimio.reader.__SWIFTNamedColumnDataset` instances are also
    wrapped, using a :class:`~swiftgalaxy.reader._SWIFTNamedColumnDatasetHelper` class.

    For an overview of available features see the examples below, and the
    narrative documentation pages.

    Parameters
    ----------
    snapshot_filename : :obj:`str`
        Name of file containing snapshot.

    halo_catalogue : :class:`~swiftgalaxy.halo_catalogues._HaloCatalogue` (optional), \
    default: ``None``
        A halo catalogue instance from :mod:`swiftgalaxy.halo_catalogues`, e.g. a
        :class:`swiftgalaxy.halo_catalogues.SOAP` instance.

    auto_recentre : :obj:`bool` (optional), default: ``True``
        If ``True``, the coordinate system will be automatically recentred on
        the position and velocity centres defined by the ``halo_catalogue``.

    transforms_like_coordinates : :obj:`set` (optional), default: ``set()``
        Names of fields that behave as spatial coordinates. It is assumed that
        these exist for all present particle types. When the coordinate system
        is rotated or translated, the associated arrays will be transformed
        accordingly. The ``coordinates`` dataset (or its alternative name given
        in the ``coordinates_dataset_name`` parameter) is implicitly assumed to
        behave as spatial coordinates.

    transforms_like_velocities : :obj:`set` (optional), default: ``set()``
        Names of fields that behave as velocities. It is assumed that these
        exist for all present particle types. When the coordinate system is
        rotated or boosted, the associated arrays will be transformed
        accordingly. The ``velocities`` dataset (or its alternative name given
        in the ``velocities_dataset_name`` parameter) is implicitly assumed to
        behave as velocities.

    id_particle_dataset_name : :obj:`str` (optional), default: ``"particle_ids"``
        Name of the dataset containing the particle IDs, assumed to be the same
        for all present particle types.

    coordinates_dataset_name : :obj:`str` (optional), default: ``"velocities"``
        Name of the dataset containing the particle spatial coordinates,
        assumed to be the same for all present particle types.

    velocities_dataset_name : :obj:`str` (optional), default: ``"velocities"``
        Name of the dataset containing the particle velocities, assumed to be
        the same for all present particle types.

    coordinate_frame_from : :class:`~swiftgalaxy.reader.SWIFTGalaxy` (optional), \
    default: ``None``
        Another :class:`~swiftgalaxy.reader.SWIFTGalaxy` to copy the coordinate frame
        (centre and rotation) and velocity coordinate frame (boost and rotation) from.

    See Also
    --------
    :class:`_SWIFTGroupDatasetHelper`
    :class:`_SWIFTNamedColumnDatasetHelper`
    :mod:`swiftgalaxy.halo_catalogues`

    Examples
    --------

    Assuming we have a snapshot file :file:`{snap}.hdf5`, and velociraptor
    outputs :file:`{halos}.properties`, :file:`{halos}.catalog_groups`, etc.,
    with the default names for coordinates, velocities and particle_ids, we can
    initialise a :class:`SWIFTGalaxy` for the first row (indexed from 0) in the
    halo catalogue very easily:

    ::

        from swiftgalaxy import SWIFTGalaxy, Velociraptor
        mygalaxy = SWIFTGalaxy(
            'snap.hdf5',
            Velociraptor(
                'halos',
                halo_index=0
            )
        )

    Like a :class:`~swiftsimio.reader.SWIFTDataset`, the particle datasets are
    accessed as below, and all data are loaded 'lazily', on demand.

    ::

        mygalaxy.gas.particle_ids
        mygalaxy.dark_matter.coordinates

    However, information from the halo catalogue is used to select only the
    particles identified as bound to this galaxy. The coordinate system is
    centred in both position and velocity on the centre and peculiar velocity
    of the galaxy, as determined by the halo catalogue. The coordinate system can
    be further manipulated, and all particle arrays will stay in a consistent
    reference frame at all times.

    Again like for a :class:`~swiftsimio.reader.SWIFTDataset`, the units and
    metadata are available:

    ::

        mygalaxy.units
        mygalaxy.metadata

    The halo catalogue interface is accessible as shown below. What this interface
    looks like depends on the halo catalogue being used, but will provide values
    for the individual galaxy of interest.

    ::

        mygalaxy.halo_catalogue

    In this case with :class:`~swiftgalaxy.halo_catalogues.Velociraptor`, we can
    get the virial mass like this:

    ::

        mygalaxy.halo_catalogue.masses.mvir

    For a complete description of available features see the narrative
    documentation pages.
    """

    snapshot_filename: str
    halo_catalogue: Optional[_HaloCatalogue]
    transforms_like_coordinates: Set[str]
    transforms_like_velocities: Set[str]
    id_particle_dataset_name: str
    coordinates_dataset_name: str
    velocities_dataset_name: str
    _initialized: bool
    _spatial_mask: SWIFTMask
    _extra_mask: Optional[MaskCollection]
    _warn_on_read: bool

    def __init__(
        self,
        snapshot_filename: str,
        halo_catalogue: Optional[_HaloCatalogue],
        auto_recentre: bool = True,
        transforms_like_coordinates: Set[str] = set(),
        transforms_like_velocities: Set[str] = set(),
        id_particle_dataset_name: str = "particle_ids",
        coordinates_dataset_name: str = "coordinates",
        velocities_dataset_name: str = "velocities",
        coordinate_frame_from: Optional["SWIFTGalaxy"] = None,
    ):
        self._particle_dataset_helpers = dict()
        self.snapshot_filename = snapshot_filename
        self.halo_catalogue = halo_catalogue
        self.transforms_like_coordinates = {coordinates_dataset_name}.union(
            transforms_like_coordinates
        )
        self.transforms_like_velocities = {velocities_dataset_name}.union(
            transforms_like_velocities
        )
        self.id_particle_dataset_name = id_particle_dataset_name
        self.coordinates_dataset_name = coordinates_dataset_name
        self.velocities_dataset_name = velocities_dataset_name
        if not hasattr(self, "_coordinate_like_transform"):
            self._coordinate_like_transform = np.eye(4)
        if not hasattr(self, "_velocity_like_transform"):
            self._velocity_like_transform = np.eye(4)
        if self.halo_catalogue is None:
            # in server mode we don't have a halo_catalogue yet
            pass
        elif self.halo_catalogue._user_spatial_offsets is not None:
            self._spatial_mask = self.halo_catalogue._get_user_spatial_mask(
                self.snapshot_filename
            )
        else:
            self._spatial_mask = self.halo_catalogue._get_spatial_mask(
                self.snapshot_filename
            )
        super().__init__(snapshot_filename, mask=self._spatial_mask)
        if auto_recentre is True and coordinate_frame_from is not None:
            raise ValueError(
                "Cannot use coordinate_frame_from with auto_recentre=True."
            )
        elif coordinate_frame_from is not None:
            if (
                coordinate_frame_from.metadata.units.length
                != self.metadata.units.length
            ) or (
                coordinate_frame_from.metadata.units.time != self.metadata.units.time
            ):
                raise ValueError(
                    "Internal units (length and time) of coordinate_frame_from don't"
                    " match."
                )
            self._coordinate_like_transform = (
                coordinate_frame_from._coordinate_like_transform
            )
            self._velocity_like_transform = (
                coordinate_frame_from._velocity_like_transform
            )
        for particle_name in self.metadata.present_group_names:
            # We'll make a custom type to present a nice name to the user.
            nice_name = swiftsimio_metadata.particle_types.particle_name_class[
                getattr(self.metadata, f"{particle_name}_properties").group
            ]
            TypeDatasetHelper = type(
                f"{nice_name}DatasetHelper", (_SWIFTGroupDatasetHelper, object), dict()
            )
            self._particle_dataset_helpers[particle_name] = TypeDatasetHelper(
                super().__getattribute__(particle_name), self
            )

        if not hasattr(self, "_extra_mask"):
            self._extra_mask = None
        if (
            self.halo_catalogue is not None
        ):  # in server mode we don't have a halo_catalogue yet
            self._extra_mask = self.halo_catalogue._get_extra_mask(self)
        if self._extra_mask is not None:
            # need to mask any already loaded data
            for particle_name in self.metadata.present_group_names:
                if getattr(self._extra_mask, particle_name) is None:
                    continue
                particle_metadata = getattr(
                    self.metadata, f"{particle_name}_properties"
                )
                for field_name in particle_metadata.field_names:
                    if getattr(self, particle_name)._is_namedcolumns(field_name):
                        named_columns_dataset = getattr(
                            getattr(self, particle_name), f"{field_name}"
                        )._named_column_dataset
                        for column in named_columns_dataset.named_columns:
                            data = getattr(named_columns_dataset, f"_{column}")
                            if data is None:
                                continue
                            setattr(
                                named_columns_dataset,
                                f"_{column}",
                                data[getattr(self._extra_mask, particle_name)],
                            )
                    else:
                        data = getattr(getattr(self, particle_name), f"_{field_name}")
                        if data is None:
                            continue
                        setattr(
                            # bypass helper:
                            super().__getattribute__(particle_name),
                            f"_{field_name}",
                            data[getattr(self._extra_mask, particle_name)],
                        )
        else:
            self._extra_mask = MaskCollection(
                **{k: None for k in self.metadata.present_group_names}
            )

        if auto_recentre and self.halo_catalogue is not None:
            self.recentre(self.halo_catalogue.centre)
            self.recentre_velocity(self.halo_catalogue.velocity_centre)

        self._warn_on_read = False
        # probably better to set False above and True here, then check value rather
        # than existence elsewhere:
        self._initialised = True

        return

    @classmethod
    def _copyinit(
        cls,
        snapshot_filename: str,
        halo_catalogue: Optional[_HaloCatalogue],
        auto_recentre: bool = True,
        transforms_like_coordinates: Set[str] = set(),
        transforms_like_velocities: Set[str] = set(),
        id_particle_dataset_name: str = "particle_ids",
        coordinates_dataset_name: str = "coordinates",
        velocities_dataset_name: str = "velocities",
        coordinate_frame_from: Optional["SWIFTGalaxy"] = None,
        _spatial_mask: Optional[SWIFTMask] = None,
        _extra_mask: Optional[MaskCollection] = None,
        _coordinate_like_transform: Optional[np.ndarray] = None,
        _velocity_like_transform: Optional[np.ndarray] = None,
        _warn_on_read: bool = False,
    ):
        """
        For internal use in copying a :class:`SWIFTGalaxy`.

        An init method with some extra parameters to facilitate copying.

        Parameters
        ----------
        snapshot_filename : :obj:`str`
            Name of file containing snapshot.

        halo_catalogue : :class:`~swiftgalaxy.halo_catalogues._HaloCatalogue` \
        (optional), default: ``None``
            A halo_catalogue instance from :mod:`swiftgalaxy.halo_catalogues`, e.g. a
            :class:`swiftgalaxy.halo_catalogues.SOAP` instance.

        auto_recentre : :obj:`bool`, default: ``True``
            If ``True``, the coordinate system will be automatically recentred on
            the position *and* velocity centres defined by the ``halo_catalogue``.

        transforms_like_coordinates : :obj:`set` containing :obj:`str`s, \
        default: ``set()``
            Names of fields that behave as spatial coordinates. It is assumed that
            these exist for all present particle types. When the coordinate system
            is rotated or translated, the associated arrays will be transformed
            accordingly. The ``coordinates`` dataset (or its alternative name given
            in the ``coordinates_dataset_name`` parameter) is implicitly assumed to
            behave as spatial coordinates.

        transforms_like_velocities : :obj:`set` containing :obj:`str`s, \
        default: ``set()``
            Names of fields that behave as velocities. It is assumed that these
            exist for all present particle types. When the coordinate system is
            rotated or boosted, the associated arrays will be transformed
            accordingly. The ``velocities`` dataset (or its alternative name given
            in the ``velocities_dataset_name`` parameter) is implicitly assumed to
            behave as velocities.

        id_particle_dataset_name : :obj:`str`, default: ``'particle_ids'``
            Name of the dataset containing the particle IDs, assumed to be the same
            for all present particle types.

        coordinates_dataset_name : :obj:`str`, default: ``'velocities'``
            Name of the dataset containing the particle spatial coordinates,
            assumed to be the same for all present particle types.

        velocities_dataset_name : :obj:`str`, default: ``'velocities'``
            Name of the dataset containing the particle velocities, assumed to be
            the same for all present particle types.

        coordinate_frame_from : :class:`~swiftgalaxy.reader.SWIFTGalaxy` (optional), \
        default: ``None``
            Another :class:`~swiftgalaxy.reader.SWIFTGalaxy` to copy the coordinate frame
            (centre and rotation) and velocity coordinate frame (boost and rotation) from.

        _spatial_mask : :class:`~swiftsimio.masks.SWIFTMask` (optional), default: ``None``
            Directly set the spatial mask (intended for internal use only).

        _extra_mask : :class:`~swiftgalaxy.masks.MaskCollection` (optional), \
        default: ``None``
            Directly set the extra mask (intended for internal use only).

        _coordinate_like_transform : :class:`~numpy.ndarray` (optional), default: ``None``
            Directly set the internal representation of the coordinate frame translations
            and rotations (intended for internal use only).

        _velocity_like_transform : :class:`~numpy.ndarray` (optional), default: ``None``
            Directly set the internal representation of the velocity frame boosts and
            rotations (intended for internal use only).

        _warn_on_read : :obj:`bool` (optional), default: ``False``
            If ``True``, warn the user when data is read from disk (e.g. should have been
            included in ``preload`` when using ``SWIFTGalaxies``.
        """
        sg = cls.__new__(cls)
        sg._spatial_mask = _spatial_mask
        sg._extra_mask = _extra_mask
        if _coordinate_like_transform is not None:
            sg._coordinate_like_transform = _coordinate_like_transform
        if _velocity_like_transform is not None:
            sg._velocity_like_transform = _velocity_like_transform
        sg._warn_on_read = _warn_on_read
        SWIFTGalaxy.__init__(
            sg,
            snapshot_filename,
            halo_catalogue,
            auto_recentre=auto_recentre,
            transforms_like_coordinates=transforms_like_coordinates,
            transforms_like_velocities=transforms_like_velocities,
            id_particle_dataset_name=id_particle_dataset_name,
            coordinates_dataset_name=coordinates_dataset_name,
            velocities_dataset_name=velocities_dataset_name,
            coordinate_frame_from=coordinate_frame_from,
        )
        sg._warn_on_read = _warn_on_read  # was set False in __init__ so do this after
        return sg

    def __str__(self) -> str:
        """
        Get a string representation of the object (noting location of the snapshot file).

        Returns
        -------
        out : :obj:`str`
            The string representation.
        """
        return f"SWIFTGalaxy at {self.snapshot_filename}."

    def __repr__(self) -> str:
        """
        Get a string representation of the object (noting location of the snapshot file).

        Returns
        -------
        out : :obj:`str`
            The string representation.
        """
        return self.__str__()

    def __getitem__(self, mask_collection: MaskCollection) -> "SWIFTGalaxy":
        """
        Apply a mask to the :class:`~swiftgalaxy.reader.SWIFTGalaxy` with square-bracket
        notation.

        To ensure internal consistency this requires producing a full ("deep") copy of
        the :class:`~swiftgalaxy.reader.SWIFTGalaxy` object and all of its contents (but
        data is masked at copy time to avoid unnecessary memory overhead).

        Parameters
        ----------
        mask_collection : :class:`swiftgalaxy.masks.MaskCollection`
            The mask to apply to the :class:`~swiftgalaxy.reader.SWIFTGalaxy`.
        """
        return self._data_copy(mask_collection=mask_collection)

    def __copy__(self) -> "SWIFTGalaxy":
        """
        Create a copy of the :class:`~swiftgalaxy.reader.SWIFTGalaxy` without copying
        data (a "shallow" copy).

        Returns
        -------
        out : :class:`~swiftgalaxy.reader.SWIFTGalaxy`
            The copy of the :class:`~swiftgalaxy.reader.SWIFTGalaxy` object.
        """
        sg = self._copyinit(
            self.snapshot_filename,
            self.halo_catalogue,
            auto_recentre=False,  # transforms overwritten below
            transforms_like_coordinates=self.transforms_like_coordinates,
            transforms_like_velocities=self.transforms_like_velocities,
            id_particle_dataset_name=self.id_particle_dataset_name,
            coordinates_dataset_name=self.coordinates_dataset_name,
            velocities_dataset_name=self.velocities_dataset_name,
            _spatial_mask=self._spatial_mask,
            _extra_mask=self._extra_mask,
            _coordinate_like_transform=self._coordinate_like_transform,
            _velocity_like_transform=self._velocity_like_transform,
            _warn_on_read=self._warn_on_read,
        )
        return sg

    def __deepcopy__(self, memo: Optional[dict] = None) -> "SWIFTGalaxy":
        """
        Create a copy of the :class:`~swiftgalaxy.reader.SWIFTGalaxy` including copying
        data (a "deep" copy).

        Parameters
        ----------
        memo : :obj:`dict` (optional), default: ``None``
            For the copy operation to keep a record of already copied objects.

        Returns
        -------
        out : :class:`~swiftgalaxy.reader.SWIFTGalaxy`
            The copy of the :class:`~swiftgalaxy.reader.SWIFTGalaxy` object.
        """
        return self._data_copy()

    def _data_copy(
        self, mask_collection: Optional[MaskCollection] = None
    ) -> "SWIFTGalaxy":
        """
        Create a copy of the :class:`~swiftgalaxy.reader.SWIFTGalaxy` including copying
        data (a "deep" copy).

        Parameters
        ----------
        mask_collection : :class:`~swiftgalaxy.masks.MaskCollection` (optional), \
        default: ``None``
            Copy only the subset of the data corresponding to the ``mask_collection``.

        Returns
        -------
        out : :class:`~swiftgalaxy.reader.SWIFTGalaxy`
            A (possibly masked) copy of the :class:`~swiftgalaxy.reader.SWIFTGalaxy`
            object.
        """
        sg = self._copyinit(
            deepcopy(self.snapshot_filename),
            deepcopy(self.halo_catalogue),
            auto_recentre=False,  # transforms overwritten below
            transforms_like_coordinates=deepcopy(self.transforms_like_coordinates),
            transforms_like_velocities=deepcopy(self.transforms_like_velocities),
            id_particle_dataset_name=deepcopy(self.id_particle_dataset_name),
            coordinates_dataset_name=deepcopy(self.coordinates_dataset_name),
            velocities_dataset_name=deepcopy(self.velocities_dataset_name),
            _spatial_mask=self._spatial_mask,
            _extra_mask=deepcopy(self._extra_mask),
            _coordinate_like_transform=deepcopy(self._coordinate_like_transform),
            _velocity_like_transform=deepcopy(self._velocity_like_transform),
            _warn_on_read=deepcopy(self._warn_on_read),
        )
        for particle_name in sg.metadata.present_group_names:
            particle_metadata = getattr(sg.metadata, f"{particle_name}_properties")
            particle_dataset_helper = getattr(self, particle_name)
            new_particle_dataset_helper = getattr(sg, particle_name)
            if mask_collection is not None:
                mask = getattr(mask_collection, particle_name)
                if mask is None:
                    mask = Ellipsis
            else:
                mask = Ellipsis
            getattr(sg, particle_name)._mask_dataset(mask)
            for field_name in particle_metadata.field_names:
                if particle_dataset_helper._is_namedcolumns(field_name):
                    named_columns_helper = getattr(particle_dataset_helper, field_name)
                    new_named_columns_helper = getattr(
                        new_particle_dataset_helper, field_name
                    )
                    for named_column in named_columns_helper.named_columns:
                        data = getattr(named_columns_helper, f"_{named_column}")
                        if data is not None:
                            setattr(
                                new_named_columns_helper, f"_{named_column}", data[mask]
                            )
                else:
                    data = getattr(particle_dataset_helper, f"_{field_name}")
                    if data is not None:
                        setattr(
                            new_particle_dataset_helper, f"_{field_name}", data[mask]
                        )
            # cartesian_coordinates return a reference to coordinates on-the-fly:
            # no need to initialise here.
            if particle_dataset_helper._spherical_coordinates is not None:
                new_particle_dataset_helper._spherical_coordinates = dict()
                for c in ("_r", "_theta", "_phi"):
                    new_particle_dataset_helper._spherical_coordinates[c] = (
                        particle_dataset_helper._spherical_coordinates[c][mask]
                    )
            if particle_dataset_helper._spherical_velocities is not None:
                new_particle_dataset_helper._spherical_velocities = dict()
                for c in ("_v_r", "_v_t", "_v_p"):
                    new_particle_dataset_helper._spherical_velocities[c] = (
                        particle_dataset_helper._spherical_velocities[c][mask]
                    )
            if particle_dataset_helper._cylindrical_coordinates is not None:
                new_particle_dataset_helper._cylindrical_coordinates = dict()
                for c in ("_rho", "_phi", "_z"):
                    new_particle_dataset_helper._cylindrical_coordinates[c] = (
                        particle_dataset_helper._cylindrical_coordinates[c][mask]
                    )
            if particle_dataset_helper._cylindrical_velocities is not None:
                new_particle_dataset_helper._cylindrical_velocities = dict()
                for c in ("_v_rho", "_v_phi", "_v_z"):
                    new_particle_dataset_helper._cylindrical_velocities[c] = (
                        particle_dataset_helper._cylindrical_velocities[c][mask]
                    )
        return sg

    def __getattribute__(self, attr: str) -> Any:
        """
        Get an attribute of the object.

        Redirects requests for datasets to the corresponding :mod:`swiftgalaxy` helper
        classes.

        Parameters
        ----------
        attr : :obj:`str`
            The name of the attribute.

        Returns
        -------
        out : :obj:`object`
            The value of the attribute.
        """
        # __getattr__ is only checked if the attribute is not found
        # __getattribute__ is checked promptly
        # Note always use super().__getattribute__(...)
        # or object.__getattribute__(self, ...) as appropriate
        # to avoid infinite recursion.
        try:
            metadata = super().__getattribute__("metadata")
        except AttributeError:
            # guard against accessing metadata before it is loaded
            return super().__getattribute__(attr)
        else:
            if attr in metadata.present_group_names:
                # We are entering a <ParticleType>Dataset, return helper.
                return object.__getattribute__(self, "_particle_dataset_helpers")[attr]
            else:
                return super().__getattribute__(attr)

    def __setattr__(self, attr: str, value: Any) -> None:
        """
        Set an attribute of the object.

        Redirects setting attributes to the helper objects when relevant.

        Parameters
        ----------
        attr : :obj:`str`
            Name of the attribute.
        value : :obj:`object`
            The value to assign to the attribute.
        """
        if (not hasattr(self, "_initialised")) or (
            attr not in self.metadata.present_group_names
        ):
            object.__setattr__(self, attr, value)
        else:
            # attr in self.metadata.present_group_names
            self._particle_dataset_helpers[attr] = value
        return

    def rotate(self, rotation: Rotation) -> None:
        """
        Apply a rotation to the particle spatial coordinates.

        The provided rotation is applied to all particle coordinates. All
        datasets specified in the ``transforms_like_coordinates`` and
        ``transforms_like_velocities`` arguments to
        :class:`SWIFTGalaxy` are transformed (by default
        ``coordinates`` and ``velocities`` for all present particle types).

        Parameters
        ----------
        rotation : :class:`scipy.spatial.transform.Rotation`
            The rotation to be applied.
            :class:`~scipy.spatial.transform.Rotation` supports several input
            formats, including axis-angle, rotation matrices, and others.
        """
        rotation_matrix = rotation.as_matrix()
        rotatable = self.transforms_like_coordinates | self.transforms_like_velocities
        for particle_name in self.metadata.present_group_names:
            dataset = getattr(self, particle_name)._particle_dataset
            for field_name in rotatable:
                field_data = getattr(dataset, f"_{field_name}")
                if field_data is not None:
                    field_data = _apply_rotmat(field_data, rotation_matrix)
                    setattr(dataset, f"_{field_name}", field_data)
        rotmat4 = np.eye(4)
        rotmat4[:3, :3] = rotation_matrix
        self._append_to_coordinate_like_transform(rotmat4)
        self._append_to_velocity_like_transform(rotmat4)
        self.wrap_box()
        return

    def _transform(self, transform4: cosmo_array, boost: bool = False) -> None:
        """
        Apply a 4x4 transformation matrix to either the spatial or velocity coordinates.

        For internal use, users should use :meth:`translate`, :meth:`boost` or
        :meth:`rotate` methods as approprirate instead.

        Parameters
        ----------
        transform4 : :class:`~numpy.ndarray`
            The transformation to be applied.
        boost : :obj:`bool`
            If ``True``, translate the velocity coordinates, else translate the spatial
            coordinates.
        """
        # assumes that the input transformation has compatible implicit units, so not
        # intended for users
        transformable = (
            self.transforms_like_velocities
            if boost
            else self.transforms_like_coordinates
        )
        for particle_name in self.metadata.present_group_names:
            dataset = getattr(self, particle_name)._particle_dataset
            for field_name in transformable:
                field_data = getattr(dataset, f"_{field_name}")
                if field_data is not None:
                    field_data = _apply_4transform(
                        field_data, transform4.to_value(), transform4.units
                    )
                    setattr(dataset, f"_{field_name}", field_data)
        if boost:
            self._append_to_velocity_like_transform(transform4)
        else:
            self._append_to_coordinate_like_transform(transform4)
        if not boost:
            self.wrap_box()

    def _translate(self, translation: cosmo_array, boost: bool = False) -> None:
        """
        Apply a translation to either the spatial or velocity coordinates.

        For internal use, users should use :meth:`translate` or :meth:`boost` as
        approprirate instead.

        Parameters
        ----------
        translation : :class:`~swiftsimio.objects.cosmo_array`
            The translation to be applied.
        boost : :obj:`bool`
            If ``True``, translate the velocity coordinates, else translate the spatial
            coordinates.
        """
        translatable = (
            self.transforms_like_velocities
            if boost
            else self.transforms_like_coordinates
        )
        for particle_name in self.metadata.present_group_names:
            dataset = getattr(self, particle_name)._particle_dataset
            for field_name in translatable:
                field_data = getattr(dataset, f"_{field_name}")
                if field_data is not None:
                    field_data = _apply_translation(field_data, translation)
                    setattr(dataset, f"_{field_name}", field_data)
        if boost:
            transform_units = self.metadata.units.length / self.metadata.units.time
        else:
            transform_units = self.metadata.units.length
        transform4 = np.eye(4)
        if hasattr(translation, "comoving"):
            transform4[3, :3] = translation.to_comoving().to_value(transform_units)
        else:
            transform4[3, :3] = translation.to_value(transform_units)
            warn(
                "Translation assumed to be in comoving (not physical) coordinates.",
                category=RuntimeWarning,
            )
        if boost:
            self._append_to_velocity_like_transform(transform4)
        else:
            self._append_to_coordinate_like_transform(transform4)
        if not boost:
            self.wrap_box()
        return

    @property
    def centre(self) -> cosmo_array:
        """
        The current origin of the coordinate reference frame with respect to the native
        simulation coordinate reference frame.

        Returns
        -------
        out : :class:`~swiftsimio.objects.cosmo_array`
            The origin of the coordinate reference frame.
        """
        transform_units = self.metadata.units.length
        transform = np.linalg.inv(self._coordinate_like_transform)
        return _apply_4transform(
            cosmo_array(
                np.zeros((1, 3)),
                units=transform_units,
                comoving=True,
                cosmo_factor=cosmo_factor(
                    a**1, scale_factor=self.metadata.scale_factor
                ),
            ),
            transform,
            transform_units,
        ).squeeze()

    @property
    def velocity_centre(self) -> cosmo_array:
        """
        The current origin of the velocity reference frame with respect to the native
        simulation velocity reference frame.

        Returns
        -------
        out : :class:`~swiftsimio.objects.cosmo_array`
            The origin of the velocity reference frame.
        """
        transform_units = self.metadata.units.length / self.metadata.units.time
        transform = np.linalg.inv(self._velocity_like_transform)
        return _apply_4transform(
            cosmo_array(
                np.zeros((1, 3)),
                units=transform_units,
                comoving=True,
                cosmo_factor=cosmo_factor(
                    a**0, scale_factor=self.metadata.scale_factor
                ),
            ),
            transform,
            transform_units,
        ).squeeze()

    @property
    def rotation(self) -> Rotation:
        """
        The current rotation of the coordinate frame.

        Returns
        -------
        out : :class:`scipy.spatial.transform.Rotation`
            The current rotation.
        """
        return Rotation.from_matrix(self._coordinate_like_transform[:3, :3])

    def translate(self, translation: cosmo_array) -> None:
        """
        Apply a translation to the particle spatial coordinates.

        The provided translation vector is added to all particle coordinates.
        If the new centre position that should be set to zero is known, use
        :meth:`recentre` instead. All datasets
        specified in the ``transforms_like_coordinates`` argument to
        :class:`SWIFTGalaxy` are transformed (by default
        ``coordinates`` for all present particle types).

        Parameters
        ----------
        translation : :class:`~swiftsimio.objects.cosmo_array`
            The vector to translate by.

        See Also
        --------
        :meth:`recentre`
        """
        self._translate(translation)
        return

    def boost(self, boost: cosmo_array) -> None:
        """
        Apply a 'boost' to the velocity coordinates.

        The provided velocity vector is added to all particle velocities. If
        the 'reference velocity' that should be set to zero is known, use
        :meth:`recentre_velocity` instead. All
        datasets specified in the ``transforms_like_velocities`` argument to
        :class:`SWIFTGalaxy` are transformed (by default
        ``velocities`` for all present particle types).

        Parameters
        ----------
        boost : :class:`~swiftsimio.objects.cosmo_array`
            The velocity to boost by.

        See Also
        --------
        :meth:`recentre_velocity`
        """
        self._translate(boost, boost=True)
        return

    def recentre(self, new_centre: cosmo_array) -> None:
        """
        Set a new centre for the particle spatial coordinates.

        The provided (spatial) coordinate centre is set to zero by subtracting
        it from the particle coordinates. Note that this is the new centre in
        the current coordinate system (not e.g. the simulation box
        coordinates). If the coordinate offset to be applied is known, use
        :meth:`translate` instead. All datasets
        specified in the ``transforms_like_coordinates`` argument to
        :class:`SWIFTGalaxy` are transformed (by default
        ``coordinates`` for all present particle types).

        Parameters
        ----------
        new_centre : :class:`~swiftsimio.objects.cosmo_array`
            The new centre for the (spatial) coordinate system.

        See Also
        --------
        :meth:`translate`
        """
        self._translate(-new_centre)
        return

    def recentre_velocity(self, new_centre: cosmo_array) -> None:
        """
        Recentre the velocity coordinates.

        The provided velocity coordinate is set to zero by subtracting it from
        the particle velocities. Note that this is the new velocity centre in
        the current coordinate system (not e.g. the simulation box
        coordinates). If the velocity offset to be applied is known, use
        :meth:`boost` instead. All
        datasets specified in the ``transforms_like_velocities`` argument to
        :class:`SWIFTGalaxy` are transformed (by default
        ``velocities`` for all present particle types).

        Parameters
        ----------
        new_centre : :class:`~swiftsimio.objects.cosmo_array`
            The new centre for the velocity coordinate system.

        See Also
        --------
        :meth:`boost`
        """
        self._translate(-new_centre, boost=True)

    def wrap_box(self) -> None:
        """
        Wrap the particle coordinates in a periodic box.

        Recentres a particle coordinates from a periodic simulation volume such
        that the coordinate (0, 0, 0) is in the centre and the axes of the
        volume are aligned with the coordinate axes.

        .. note::
            This is invoked automatically after any coordinate translations or
            rotations, so manually calling this function should usually not be
            necessary.
        """
        for particle_name in self.metadata.present_group_names:
            dataset = getattr(self, particle_name)._particle_dataset
            for field_name in self.transforms_like_coordinates:
                field_data = getattr(dataset, f"_{field_name}")
                if field_data is not None:
                    field_data = _apply_box_wrap(field_data, self.metadata.boxsize)
                    setattr(dataset, f"_{field_name}", field_data)
        return

    def mask_particles(self, mask_collection: MaskCollection) -> None:
        """
        Select a subset of the currently selected particles.

        The masks to be applied can by in any format accepted by a
        :class:`~swiftsimio.objects.cosmo_array` via
        :meth:`~swiftsimio.objects.cosmo_array.__getitem__` and should be
        collected in a :class:`swiftgalaxy.masks.MaskCollection`. The
        selection is applied permanently to all particle datasets for this
        galaxy. Temporary masks (e.g. for interactive use) can be created by
        using the :meth:`~SWIFTGalaxy.__getitem__` (square brackets) method of
        the :class:`SWIFTGalaxy`, any of its associated
        :class:`_SWIFTGroupDatasetHelper` or
        :class:`_SWIFTNamedColumnDatasetHelper` objects, but
        note that to ensure internal consistency, these return a masked copy of
        the *entire* :class:`SWIFTGalaxy`, and are therefore
        relatively memory-intensive. Masking individual
        :class:`~swiftsimio.objects.cosmo_array` datasets with
        :meth:`~swiftsimio.objects.cosmo_array.__getitem__`
        avoids this: only masked copies of the individual arrays are returned
        in this case.

        Parameters
        ----------
        mask_collection : :class:`swiftgalaxy.masks.MaskCollection`
            Set of masks to be applied to each particle type. Particle types
            may be omitted by setting their mask to None, or simply omitting
            them from the :class:`swiftgalaxy.masks.MaskCollection`.
        """
        for particle_name in self.metadata.present_group_names:
            mask = getattr(mask_collection, particle_name)
            if mask is not None:
                getattr(self, particle_name)._mask_dataset(mask)
        return

    def _append_to_coordinate_like_transform(self, transform: np.ndarray) -> None:
        """
        Add a new transformation to the sequence of transformations for the spatial-like
        coordinates.

        The cumulative transformation is stored as a single 4x4 transformation matrix,
        so we update the current transformation using a dot product. This voids any
        derived (spherical/cylindrical) coordinates.

        Parameters
        ----------
        transform : :class:`~numpy.ndarray`
            The transform to add to the cumulative coordinate transformation.
        """
        self._coordinate_like_transform = self._coordinate_like_transform.dot(transform)
        self._void_derived_coordinates()
        return

    def _append_to_velocity_like_transform(self, transform: np.ndarray) -> None:
        """
        Add a new transformation to the sequence of transformations for the velocity-like
        coordinates.

        The cumulative transformation is stored as a single 4x4 transformation matrix,
        so we update the current transformation using a dot product. This voids any
        derived (spherical/cylindrical) coordinates.

        Parameters
        ----------
        transform : :class:`~numpy.ndarray`
            The transform to add to the cumulative velocity transformation.
        """
        self._velocity_like_transform = self._velocity_like_transform.dot(transform)
        self._void_derived_coordinates()
        return

    def _void_derived_coordinates(self) -> None:
        """
        Reset internal references to spherical/cylindrical coordinates to None (e.g.
        because they are no longer valid).
        """
        # Transforming implies conversion back to cartesian, it's therefore
        # cheaper to just delete any non-cartesian coordinates when a
        # transform occurs and lazily re-calculate them as needed.
        for particle_name in self.metadata.present_group_names:
            getattr(self, particle_name)._void_derived_coordinates()
        return



import pytest
import numpy as np
import unyt as u
from unyt.testing import assert_allclose_units
from toysnap import (
    toysnap_filename,
    ToyHF,
    present_particle_types,
    toysoap_virtual_snapshot_filename,
    toysoap_filename,
    toyvr_filebase,
    toycaesar_filename,
    create_toysoap,
    create_toyvr,
    create_toycaesar,
    remove_toysoap,
    remove_toyvr,
    remove_toycaesar,
)
from conftest import hfs
from swiftsimio.objects import cosmo_array, cosmo_factor, a
from swiftgalaxy.reader import SWIFTGalaxy
from swiftgalaxy.iterator import SWIFTGalaxies
from swiftgalaxy.halo_catalogues import Standalone, SOAP, Velociraptor, Caesar
from swiftsimio import mask


class TestSWIFTGalaxies:

    def test_eval_sparse_optimized_solution(self, toysnap):
        """
        Check that the sparse solution is chosen when optimal and matches expectations
        for a case that we can work out by hand.
        """
        # place a single target in the centre of each cell
        # this should make sparse iteration optimal
        # at a cost of 2 cell reads
        sgs = SWIFTGalaxies(
            toysnap_filename,
            Standalone(
                centre=cosmo_array(
                    [[2.5, 5.0, 5.0], [7.5, 5.0, 5.0]],
                    u.Mpc,
                    comoving=True,
                    cosmo_factor=cosmo_factor(a**1, 1.0),
                ),
                velocity_centre=cosmo_array(
                    [[0, 0, 0] * 2],
                    u.km / u.s,
                    comoving=True,
                    cosmo_factor=cosmo_factor(a**0, 1.0),
                ),
                spatial_offsets=cosmo_array(
                    [[-1.0, 1.0], [-1.0, 1.0], [-1.0, 1.0]],
                    u.Mpc,
                    comoving=True,
                    cosmo_factor=cosmo_factor(a**1, 1.0),
                ),
            ),
            preload={  # just to keep warnings quiet
                "gas.particle_ids",
                "dark_matter.particle_ids",
                "stars.particle_ids",
                "black_holes.particle_ids",
            },
        )
        sparse_solution = sgs._sparse_optimized_solution
        assert_allclose_units(
            sparse_solution["regions"],
            cosmo_array(
                np.array(
                    [
                        [[0.01, 0.99], [0.01, 0.99], [0.01, 0.99]],
                        [[1.01, 1.99], [0.01, 0.99], [0.01, 0.99]],
                    ]
                )
                * np.array([[[5], [10], [10]]]),
                u.Mpc,
                comoving=True,
                cosmo_factor=cosmo_factor(a**1, 1.0),
            ),
            atol=0.0001 * u.Mpc,
        )
        for ss, expected in zip(
            sparse_solution["region_target_indices"], [np.array([0]), np.array([1])]
        ):
            assert np.allclose(ss, expected, atol=0)
        assert sparse_solution["cost"] == 2
        for k in sgs._solution.keys():
            if isinstance(sgs._solution[k], np.integer):
                assert sgs._solution[k] == sparse_solution[k]
            elif isinstance(sgs._solution[k], list):
                for i in range(len(sgs._solution[k])):
                    assert np.allclose(
                        sgs._solution[k][i], sparse_solution[k][i], atol=0
                    )
            else:
                assert_allclose_units(
                    sgs._solution[k], sparse_solution[k], atol=0.001 * u.Mpc
                )

    def test_eval_dense_optimized_solution(self, toysnap):
        """
        Check that the dense solution is chosen when optimal and matches expectations
        for a case that we can work out by hand.
        """
        # Place a single target in the centre of each cell
        # and ones straddling many faces, vertices and corners
        # this should make dense iteration optimal
        # at a cost of 2-4 cell reads.
        # * Technically this solution is inefficient because
        # the cell regions are repeated, because the code is
        # not clever enough to recognize copies wrapped through
        # the periodic bounday. But having so few cells is
        # not an expected case for "real" simulations.
        # The dense solution could start to include some
        # copies if the target region is larger than ~0.5 the
        # box size in any dimension.
        sgs = SWIFTGalaxies(
            toysnap_filename,
            Standalone(
                centre=cosmo_array(
                    [
                        [2.5, 5.0, 5.0],
                        [5.0, 5.0, 5.0],
                        [7.5, 5.0, 5.0],
                        [5.0, 0.0, 0.0],
                        [5.0, 0.0, 9.9],
                        [5.0, 9.9, 0.0],
                        [5.0, 9.9, 9.9],
                        [9.9, 0.0, 9.9],
                        [9.9, 9.9, 0.0],
                        [9.9, 9.9, 9.9],
                    ],
                    u.Mpc,
                    comoving=True,
                    cosmo_factor=cosmo_factor(a**1, 1.0),
                ),
                velocity_centre=cosmo_array(
                    [[0, 0, 0] * 10],
                    u.km / u.s,
                    comoving=True,
                    cosmo_factor=cosmo_factor(a**0, 1.0),
                ),
                spatial_offsets=cosmo_array(
                    [[-1.0, 1.0], [-1.0, 1.0], [-1.0, 1.0]],
                    u.Mpc,
                    comoving=True,
                    cosmo_factor=cosmo_factor(a**1, 1.0),
                ),
            ),
            preload={  # just to keep warnings quiet
                "gas.particle_ids",
                "dark_matter.particle_ids",
                "stars.particle_ids",
                "black_holes.particle_ids",
            },
        )
        dense_solution = sgs._dense_optimized_solution
        assert_allclose_units(
            dense_solution["regions"],
            cosmo_array(
                np.array(
                    [
                        [[0.005, 4.995], [0.01, 9.99], [0.01, 9.99]],
                        [[5.005, 9.995], [0.01, 9.99], [0.01, 9.99]],
                        [[2.505, 7.495], [5.01, 14.99], [5.01, 14.99]],
                        [[7.505, 12.495], [5.01, 14.99], [5.01, 14.99]],
                    ]
                ),
                u.Mpc,
                comoving=True,
                cosmo_factor=cosmo_factor(a**1, 1.0),
            ),
            atol=0.001 * u.Mpc,
        )
        assert len(dense_solution["region_target_indices"]) == 4
        for ds, expected in zip(
            dense_solution["region_target_indices"],
            [
                np.array([0]),
                np.array([1, 2, 3, 4, 5, 7, 8]),
                np.array([6]),
                np.array([9]),
            ],
        ):
            assert np.allclose(ds, expected, atol=0)
        assert dense_solution["cost_min"] == 4
        assert dense_solution["cost_max"] == 32
        for k in sgs._solution.keys():
            if isinstance(sgs._solution[k], np.integer):
                assert sgs._solution[k] == dense_solution[k]
            elif isinstance(sgs._solution[k], list):
                for i in range(len(sgs._solution[k])):
                    assert np.allclose(
                        sgs._solution[k][i], dense_solution[k][i], atol=0
                    )
            else:
                assert_allclose_units(
                    sgs._solution[k], dense_solution[k], atol=0.001 * u.Mpc
                )

    def test_iteration_order(self, sgs):
        """
        Check that the iteration order agrees with that computed in the two iteration
        optimizations.
        """
        # force dense solution
        sgs._solution = sgs._dense_optimized_solution
        assert np.allclose(
            sgs.iteration_order,
            np.concatenate(sgs._dense_optimized_solution["region_target_indices"]),
            atol=0,
        )
        # force sparse solution
        sgs._solution = sgs._sparse_optimized_solution
        assert np.allclose(
            sgs.iteration_order,
            np.concatenate(sgs._sparse_optimized_solution["region_target_indices"]),
            atol=0,
        )

    def test_iterate(self, sgs):
        """
        Check that we iterate over the right number of SWIFTGalaxy objects and that
        they behave like a SWIFTGalaxy created on its own for each target.
        """
        count = 0
        for sg_from_sgs in sgs:
            sg = SWIFTGalaxy(
                toysnap_filename,
                ToyHF(
                    snapfile=toysnap_filename, index=sg_from_sgs.halo_catalogue.index
                ),
            )
            for ptype in present_particle_types.values():
                assert np.all(
                    getattr(sg_from_sgs._extra_mask, ptype)
                    == getattr(sg._extra_mask, ptype)
                )
            count += 1
        assert count == len(sgs.halo_catalogue.index)

    @pytest.mark.parametrize("extra_mask", ["bound_only", None])
    def test_preload(self, toysnap_withfof, hf_multi, extra_mask):
        """
        Make sure that data that we ask to have pre-loaded is actually pre-loaded.
        """
        hf_multi.extra_mask = extra_mask
        sgs = SWIFTGalaxies(
            (
                toysoap_virtual_snapshot_filename
                if isinstance(hf_multi, SOAP)
                else toysnap_filename
            ),
            hf_multi,
            preload={
                "gas.particle_ids",
                "dark_matter.particle_ids",
                "stars.particle_ids",
                "black_holes.particle_ids",
            },
        )
        region_mask = mask(toysnap_filename)
        region_mask.constrain_spatial(sgs._solution["regions"][0])
        sgs._start_server(region_mask)
        sgs._preload()
        for preload_field in sgs.halo_catalogue._get_preload_fields(sgs._server):
            ptype, field = preload_field.split(".")
            assert (
                getattr(getattr(sgs._server, ptype)._particle_dataset, f"_{field}")
                is not None
            )

    def test_warn_on_no_preload(self, toysnap):
        """
        Check that we warn users if they don't specify anything to pre-load since this
        probably indicates that they're using the SWIFTGalaxies class inefficiently.
        """
        with pytest.warns(RuntimeWarning, match="No data specified to preload"):
            SWIFTGalaxies(
                toysnap_filename,
                ToyHF(index=[0, 1]),
            )

    def test_warn_on_read_not_preloaded(self, sgs):
        """
        Check that we warn users when data is loaded while iterating over a SWIFTGalaxies
        since this probably indicates that they're using the class inefficiently.
        """
        assert "gas.coordinates" not in sgs.preload
        for sg in sgs:
            with pytest.warns(RuntimeWarning, match="should it be preloaded"):
                sg.gas.coordinates

    def test_exception_on_repeated_targets(self, toysnap):
        """
        Due to especially swiftsimio's masking behaviour having duplicate targets in the
        list for a SWIFTGalaxies causes all kinds of problems, so make sure we raise an
        exception if a user tries to do this.
        """
        with pytest.raises(ValueError, match="must not contain duplicates"):
            SWIFTGalaxies(
                toysnap_filename,
                ToyHF(index=[0, 0]),
            )

    def test_map(self, toysnap_withfof, hf_multi):
        """
        Check that the map method returns results in the same order as the input target
        list. We're careful in this test to make sure that the iteration order is
        different from the input list order.
        """
        sgs = SWIFTGalaxies(
            (
                toysoap_virtual_snapshot_filename
                if isinstance(hf_multi, SOAP)
                else toysnap_filename
            ),
            hf_multi,
            preload={  # just to keep warnings quiet
                "gas.particle_ids",
                "dark_matter.particle_ids",
                "stars.particle_ids",
                "black_holes.particle_ids",
            },
        )

        def f(sg):
            if isinstance(hf_multi, Standalone):
                return int(
                    np.argwhere(
                        np.all(
                            sg.halo_catalogue.centre == sg.halo_catalogue._centre,
                            axis=1,
                        )
                    ).squeeze()
                )
            # _index_attr has leading underscore, access through property with [1:]
            return getattr(sg.halo_catalogue, sg.halo_catalogue._index_attr[1:])

        if (np.diff(sgs.iteration_order) == 1).all():
            # if we iterate in order success is trivial, ensure that we don't:
            sgs._solution["regions"] = sgs._solution["regions"][::-1]
            sgs._solution["region_target_indices"] = sgs._solution[
                "region_target_indices"
            ][::-1]
        # double-check that success won't be trivial:
        assert not (np.diff(sgs.iteration_order) == 1).all()
        # check that map returns results ordered in input order
        if isinstance(hf_multi, Standalone):
            assert sgs.map(f) == [0, 1]
            return
        assert sgs.map(f) == getattr(
            sgs.halo_catalogue, sgs.halo_catalogue._index_attr[1:]
        )

    def test_arbitrary_index_ordering(self, toysnap_withfof, hf_multi):
        """
        Check that SWIFTGalaxies gives consistent results for any order of target objects.

        Especially important for velociraptor where some logic had to be added to avoid
        hdf5 complaining about an unsorted list of indices to read from file.
        """

        def f(sg):
            if isinstance(hf_multi, Standalone):
                return int(
                    np.argwhere(
                        np.all(
                            sg.halo_catalogue.centre == sg.halo_catalogue._centre,
                            axis=1,
                        )
                    ).squeeze()
                )
            # _index_attr has leading underscore, access through property with [1:]
            return getattr(sg.halo_catalogue, sg.halo_catalogue._index_attr[1:])

        sgs_forwards = SWIFTGalaxies(
            (
                toysoap_virtual_snapshot_filename
                if isinstance(hf_multi, SOAP)
                else toysnap_filename
            ),
            hf_multi,
            preload={  # just to keep warnings quiet
                "gas.particle_ids",
                "dark_matter.particle_ids",
                "stars.particle_ids",
                "black_holes.particle_ids",
            },
        )
        map_forwards = sgs_forwards.map(f)
        if isinstance(hf_multi, Standalone):
            hf_multi._centre = hf_multi._centre[::-1]
            hf_multi._velocity_centre = hf_multi._velocity_centre[::-1]
        else:
            setattr(
                hf_multi,
                hf_multi._index_attr,
                getattr(hf_multi, hf_multi._index_attr)[::-1],
            )
        sgs_backwards = SWIFTGalaxies(
            (
                toysoap_virtual_snapshot_filename
                if isinstance(hf_multi, SOAP)
                else toysnap_filename
            ),
            hf_multi,
            preload={  # just to keep warnings quiet
                "gas.particle_ids",
                "dark_matter.particle_ids",
                "stars.particle_ids",
                "black_holes.particle_ids",
            },
        )
        map_backwards = sgs_backwards.map(f)
        if isinstance(hf_multi, Standalone):
            # because the reversed catalogue compares to the reversed catalogue in f,
            # we get the reverse-of-the-reverse in map_backwards, which should match
            # map_forwards
            assert map_forwards == map_backwards
            return
        assert map_forwards == map_backwards[::-1]

    def test_args_kwargs_to_map(self, sgs):
        """
        Make sure that we can pass extra args & kwargs to a function given to map.
        """
        extra_arg = [("foo",), ("bar",)]
        extra_kwarg = [dict(extra_kwarg="spam"), dict(extra_kwarg="eggs")]

        def f(sg, extra_arg, extra_kwarg=None):
            return extra_arg, extra_kwarg

        result = sgs.map(f, args=extra_arg, kwargs=extra_kwarg)
        assert result == [("foo", "spam"), ("bar", "eggs")]

    def test_soap_target_order_consistency(
        self, toysnap_withfof, toysoap_with_virtual_snapshot
    ):
        """
        SOAP implicitly sorts the target mask (when getting things from the catalogue
        we get them masked in the order that they appear in the catalogue, rather
        than the order that they appear in the mask - say we ask for items [1, 0]
        as a mask, we get back those two items in order [0, 1]). This test checks
        that we get a SWIFTGalaxy from a SWIFTGalaxies that matches the one
        constructed directly when the targets given to the SWIFTGalaxies are not
        in the order that they appear in the catalogue.
        """
        soaps = [
            SOAP(soap_file=toysoap_filename, soap_index=0),
            SOAP(soap_file=toysoap_filename, soap_index=1),
        ]
        soap_both = SOAP(soap_file=toysoap_filename, soap_index=[1, 0])
        sgs_individual = [
            SWIFTGalaxy(toysoap_virtual_snapshot_filename, soap) for soap in soaps
        ]
        sgs = SWIFTGalaxies(
            toysoap_virtual_snapshot_filename,
            soap_both,
            preload={"black_holes.masses"},  # just keep warnings quiet
        )
        for i, sg in enumerate(sgs):
            sg_single = sgs_individual[sgs.iteration_order[i]]
            assert_allclose_units(
                sg_single.halo_catalogue._region_centre,
                sg.halo_catalogue._region_centre,
            )
            assert_allclose_units(
                sg_single.halo_catalogue.centre,
                sg.halo_catalogue.centre,
            )
            assert_allclose_units(
                sg_single.halo_catalogue.velocity_centre,
                sg.halo_catalogue.velocity_centre,
            )
            assert_allclose_units(
                sg_single.halo_catalogue.spherical_overdensity_200_crit.centre_of_mass,
                sg.halo_catalogue.spherical_overdensity_200_crit.centre_of_mass,
            )

    @pytest.mark.parametrize("hf_type", hfs)
    def test_halo_catalogue_with_non_list_indices(self, hf_type, toysnap_withfof):
        """
        Check if we can initialize a halo_catalogue in multi-galaxy mode with
        ordered containers that are not a list.
        """
        if hf_type == "soap":
            create_toysoap(create_virtual_snapshot=True)
        elif hf_type == "vr":
            create_toyvr()
        elif "caesar" in hf_type:
            create_toycaesar()
        elif hf_type == "sa":
            return  # doesn't take an index list, nothing to test
        else:
            raise NotImplementedError  # a new halo_catalogue that we're not testing
        for init_indices in (
            np.array([0, 1], dtype=int),
            u.unyt_array([0, 1], u.dimensionless, dtype=int),
        ):
            if hf_type == "soap":
                sgs = SWIFTGalaxies(
                    toysoap_virtual_snapshot_filename,
                    SOAP(toysoap_filename, soap_index=init_indices),
                    preload={"gas.masses"},  # just to keep warnings quiet
                )
            elif hf_type == "vr":
                sgs = SWIFTGalaxies(
                    toysnap_filename,
                    Velociraptor(
                        velociraptor_filebase=toyvr_filebase, halo_index=init_indices
                    ),
                    preload={"gas.masses"},
                )
            elif hf_type == "caesar_galaxy":
                sgs = SWIFTGalaxies(
                    toysnap_filename,
                    Caesar(
                        toycaesar_filename,
                        group_type="galaxy",
                        group_index=init_indices,
                    ),
                    preload={"gas.masses"},
                )
            elif hf_type == "caesar_halo":
                sgs = SWIFTGalaxies(
                    toysnap_filename,
                    Caesar(
                        toycaesar_filename, group_type="halo", group_index=init_indices
                    ),
                    preload={"gas.masses"},
                )
            for sg in sgs:
                pass  # just go through the iteration
        if hf_type == "soap":
            remove_toysoap()
        elif hf_type == "vr":
            remove_toyvr()
        elif "caesar" in hf_type:
            remove_toycaesar()



class TestSWIFTGalaxyCreation:
    def test_sg_creation(self, sg):
        pass  # fixture created SWIFTGalaxy

    def test_soap_creation(self, soap):
        pass  # fixture created SOAP interface

    def test_vr_creation(self, vr):
        pass  # fixture created Velociraptor interface

    def test_caesar_creation(self, caesar):
        pass  # fixture created Caesar interface

    def test_sa_creation(self, sa):
        pass  # fixture created Standalone interface

    def test_sg_soap_creation(self, sg_soap):
        pass  # fixture created SWIFTGalaxy with SOAP interface

    def test_sg_vr_creation(self, sg_vr):
        pass  # fixture created SWIFTGalaxy with Velociraptor interface

    def test_sg_caesar_creation(self, sg_caesar):
        pass  # fixture created SWIFTGalaxy with Caesar interface

    def test_sg_sa_creation(self, sg_sa):
        pass  # fixture created SWIFTGalaxy with Standalone interface


class TestSWIFTGalaxiesCreation:
    def test_sgs_creation(self, sgs):
        pass  # fixture created SWIFTGalaxies

    def test_soap_multi_creation(self, soap_multi):
        pass  # fixture created SOAP interface

    def test_vr_creation(self, vr_multi):
        pass  # fixture created Velociraptor interface

    def test_caesar_creation(self, caesar_multi):
        pass  # fixture created Caesar interface

    def test_sa_creation(self, sa_multi):
        pass  # fixture created Standalone interface

    def test_sgs_soap_creation(self, sgs_soap):
        pass  # fixture created SWIFTGalaxy with SOAP interface

    def test_sgs_vr_creation(self, sgs_vr):
        pass  # fixture created SWIFTGalaxy with Velociraptor interface

    def test_sgs_caesar_creation(self, sgs_caesar):
        pass  # fixture created SWIFTGalaxy with Caesar interface

    def test_sgs_sa_creation(self, sgs_sa):
        pass  # fixture created SWIFTGalaxy with Standalone interface



"""
The :mod:`swiftsimio` masking features are optimized for spatial masking, that is
selecting regions made up of a subset of the SWIFT "top-level cells" in a simulation.
:mod:`swiftgalaxy` masking features aim to support refining these relatively coarse
spatial masks to select particles belonging to individual structures, or other
arbitrary sets of particles, and therefore needs its own masking tools. In their
long term the hope is to merge the two together, but for now the
:class:`~swiftgalaxy.masks.MaskCollection` is the recommended way to define a
selection of particles of different types for use with
:class:`~swiftgalaxy.reader.SWIFTGalaxy`.
"""

from copy import deepcopy
from typing import Optional


class MaskCollection(object):
    """
    Barebones container for mask objects.

    Takes a set of kwargs at initialisation and assigns their values to
    attributes of the object. Attempts to access a non-existent attribute
    returns :obj:`None` instead of raising an :exc:`AttributeError`.

    This is intended to hold masks that can be applied to
    :class:`~swiftsimio.objects.cosmo_array` objects under the names of
    particle types (e.g. ``gas``, ``dark_matter``, etc.), but this is not
    checked or enforced.

    Parameters
    ----------

    **kwargs
        Any items passed as kwargs will have their values passed to
        correspondingly named attributes of this object.

    Notes
    -----

    .. note::
        The :mod:`velociraptor.swift.swift` module makes some use of a
        :obj:`namedtuple` called ``MaskCollection``. These objects are not
        valid where :mod:`swiftgalaxy` functions expect a :obj:`MaskCollection`
        because :obj:`namedtuple` objects are immutable.

    Examples
    --------
    ::

        n_dm = 123  # suppose this is number of dark matter particles
        # all these masks select all particles:
        MaskCollection(
            gas=np.s_[...],
            dark_matter=np.ones(n_dm, dtype=bool),
            stars=None
        )
    """

    # Could use dataclasses module, but requires python 3.7+

    def __init__(self, **kwargs) -> None:
        for k, v in kwargs.items():
            setattr(self, k, v)
        return

    def __getattr__(self, attr: str) -> None:
        """
        Return ``None`` if an attribute of the object doesn't exist.

        This function is called if an attribute is asked for and not found.
        Instead of the usual behaviour of raising a :exc:`AttributeError`,
        ``None`` is returned.

        Parameters
        ----------
        attr : :obj:`str`
            The name of the requested attribute.

        Returns
        -------
        out : None
            If we reach calling this function the attribute is not found and we
            return ``None``.
        """
        return None

    def __copy__(self) -> "MaskCollection":
        """
        Make a copy of the :class:`~swiftgalaxy.masks.MaskCollection` without copying
        data (a "shallow" copy).

        Returns
        -------
        out : :class:`~swiftgalaxy.masks.MaskCollection`
            The copy of the :class:`~swiftgalaxy.masks.MaskCollection`.
        """
        return MaskCollection(**self.__dict__)

    def __deepcopy__(self, memo: Optional[dict] = None) -> "MaskCollection":
        """
        Make a copy of the :class:`~swiftgalaxy.masks.MaskCollection`, copying data
        (a "deep" copy).

        Parameters
        ----------
        memo : :obj:`dict` (optional), default: ``None``
            For the copy operation to keep a record of already copied objects.

        Returns
        -------
        out : :class:`~swiftgalaxy.masks.MaskCollection`
            The copy of the :class:`~swiftgalaxy.masks.MaskCollection`.
        """
        return MaskCollection(**{k: deepcopy(v) for k, v in self.__dict__.items()})



import pytest
import h5py
import numpy as np
import unyt as u
from unyt.testing import assert_allclose_units
from toysnap import (
    toysnap_filename,
    toysoap_virtual_snapshot_filename,
    toysoap_membership_filebase,
    n_g_1,
    n_g_2,
    n_g_b,
    n_g_all,
    n_dm_1,
    n_dm_b,
    n_dm_all,
    n_s_1,
    n_s_2,
    n_bh_1,
    n_bh_2,
    m_g,
    # m_dm,
    m_s,
    m_bh,
    centre_1,
    centre_2,
    vcentre_1,
    present_particle_types,
)
from swiftgalaxy import SWIFTGalaxy, MaskCollection
from swiftgalaxy.halo_catalogues import Velociraptor, Caesar, SOAP, Standalone
from swiftsimio.objects import cosmo_array, cosmo_factor, a

abstol_c = 1 * u.pc  # less than this is ~0
abstol_v = 10 * u.m / u.s  # less than this is ~0
abstol_m = 1e4 * u.Msun  # less than this is ~0
reltol_nd = 1.0e-4


class TestHaloCatalogues:
    def test_get_spatial_mask(self, hf, toysnap):
        """
        Check that we get spatial masks that we expect.
        """
        # don't use sg fixture here, just need the snapshot file
        # so don't want overhead of a SWIFTGalaxy
        spatial_mask = hf._get_spatial_mask(toysnap_filename)
        with h5py.File(toysnap_filename, "r") as snap:
            n_g_firstcell = snap["/Cells/Counts/PartType0"][0]
            n_dm_firstcell = snap["/Cells/Counts/PartType1"][0]
            n_s_firstcell = snap["/Cells/Counts/PartType4"][0]
            n_bh_firstcell = snap["/Cells/Counts/PartType5"][0]
        # We have 2 cells, covering 0-5 and 5-10 Mpc in x, and the entire range in y
        # and z. The galaxy is at (2,2,2)Mpc, so in the first cell.
        if hf._user_spatial_offsets is None:
            # all hf's except Standalone
            assert np.array_equal(spatial_mask.gas, np.array([[0, n_g_firstcell]]))
            assert np.array_equal(
                spatial_mask.dark_matter, np.array([[0, n_dm_firstcell]])
            )
            assert np.array_equal(spatial_mask.stars, np.array([[0, n_s_firstcell]]))
            assert np.array_equal(
                spatial_mask.black_holes, np.array([[0, n_bh_firstcell]])
            )
        else:
            # this is Standalone
            assert np.array_equal(
                spatial_mask.gas,
                np.array([[0, n_g_firstcell], [n_g_firstcell, n_g_all]]),
            )
            assert np.array_equal(
                spatial_mask.dark_matter,
                np.array([[0, n_dm_firstcell], [n_dm_firstcell, n_dm_all]]),
            )
            assert np.array_equal(
                spatial_mask.stars,
                np.array([[0, n_s_firstcell], [n_s_firstcell, n_s_1 + n_s_2]]),
            )
            assert np.array_equal(
                spatial_mask.black_holes,
                np.array([[0, n_bh_firstcell], [n_bh_firstcell, n_bh_1 + n_bh_2]]),
            )

    def test_get_user_spatial_mask(self, hf, toysnap):
        """
        Check that a user can override the automatic spatial mask.
        """
        # override to select both cells in the test snapshot
        hf._user_spatial_offsets = cosmo_array(
            [[-5, 5], [-5, 5], [-5, 5]],
            u.Mpc,
            comoving=True,
            cosmo_factor=cosmo_factor(a**1, 1.0),
        )
        hf.extra_mask = None
        sg = SWIFTGalaxy(toysnap_filename, hf)
        generated_spatial_mask = sg._spatial_mask
        with h5py.File(toysnap_filename, "r") as snap:
            n_g_firstcell = snap["/Cells/Counts/PartType0"][0]
            n_dm_firstcell = snap["/Cells/Counts/PartType1"][0]
            n_s_firstcell = snap["/Cells/Counts/PartType4"][0]
            n_bh_firstcell = snap["/Cells/Counts/PartType5"][0]
        assert np.array_equal(
            generated_spatial_mask.gas,
            np.array([[0, n_g_firstcell], [n_g_firstcell, n_g_all]]),
        )
        assert np.array_equal(
            generated_spatial_mask.dark_matter,
            np.array([[0, n_dm_firstcell], [n_dm_firstcell, n_dm_all]]),
        )
        assert np.array_equal(
            generated_spatial_mask.stars,
            np.array([[0, n_s_firstcell], [n_s_firstcell, n_s_1 + n_s_2]]),
        )
        assert np.array_equal(
            generated_spatial_mask.black_holes,
            np.array([[0, n_bh_firstcell], [n_bh_firstcell, n_bh_1 + n_bh_2]]),
        )

    def test_get_bound_only_extra_mask(self, hf, toysnap_withfof):
        """
        Check that bound_only extra mask has the right shape.
        """
        hf.extra_mask = "bound_only"
        if hasattr(hf, "soap_file"):
            from toysnap import soap_script
            import os

            os.system(
                f"python {soap_script} "
                f"--absolute-paths "
                f"'{toysnap_filename}' "
                f"'{toysoap_membership_filebase}."
                "{file_nr}.hdf5' "
                f"'{toysoap_virtual_snapshot_filename}' "
                "0"
            )
            sg = SWIFTGalaxy(toysoap_virtual_snapshot_filename, hf)
        else:
            try:
                sg = SWIFTGalaxy(toysnap_filename, hf)
            except NotImplementedError:
                # expected for Standalone
                return
        generated_extra_mask = sg._extra_mask
        expected_shape = dict()
        for particle_type in present_particle_types.values():
            with h5py.File(toysnap_filename, "r") as snap:
                expected_shape[particle_type] = snap[
                    "Cells/Counts/PartType"
                    f"{dict(gas=0, dark_matter=1, stars=4, black_holes=5)[particle_type]}"
                ][0]
        if hasattr(hf, "_caesar") and hf.group_type == "galaxy":
            expected_shape["dark_matter"] = None
        for particle_type in present_particle_types.values():
            if expected_shape[particle_type] is not None:
                assert (
                    getattr(generated_extra_mask, particle_type).shape
                    == expected_shape[particle_type]
                )
                assert (
                    getattr(generated_extra_mask, particle_type).sum()
                    == dict(
                        gas=n_g_1, dark_matter=n_dm_1, stars=n_s_1, black_holes=n_bh_1
                    )[particle_type]
                )

    def test_get_void_extra_mask(self, hf, toysnap):
        """
        Check that None extra mask gives expected result.
        """
        hf.extra_mask = None
        sg = SWIFTGalaxy(toysnap_filename, hf)
        generated_extra_mask = sg._extra_mask
        for particle_type in present_particle_types.values():
            assert getattr(generated_extra_mask, particle_type) is None

    def test_get_user_extra_mask(self, hf, toysnap):
        """
        Check that extra masks of different kinds have the right shape or type.
        """
        hf.extra_mask = MaskCollection(
            gas=np.r_[np.ones(100, dtype=bool), np.zeros(n_g_all - 100, dtype=bool)],
            dark_matter=None,
            stars=np.r_[np.ones(100, dtype=bool), np.zeros(n_s_1 - 100, dtype=bool)],
            black_holes=np.ones(n_bh_1, dtype=bool),
        )
        sg = SWIFTGalaxy(toysnap_filename, hf)
        generated_extra_mask = sg._extra_mask
        for particle_type in present_particle_types.values():
            if getattr(generated_extra_mask, particle_type) is None:
                assert (
                    dict(gas=100, dark_matter=None, stars=100, black_holes=n_bh_1)[
                        particle_type
                    ]
                    is None
                )
            else:
                assert (
                    getattr(generated_extra_mask, particle_type).sum()
                    == dict(gas=100, dark_matter=None, stars=100, black_holes=n_bh_1)[
                        particle_type
                    ]
                )

    def test_centre(self, hf):
        """
        Check that the _centre function returns the expected centre.
        """
        # default is minpot == 2.001 Mpc
        assert_allclose_units(
            hf.centre,
            cosmo_array(
                [centre_1 + 0.001, centre_1 + 0.001, centre_1 + 0.001],
                u.Mpc,
                comoving=True,
                cosmo_factor=cosmo_factor(a**1, 1.0),
            ),
            rtol=reltol_nd,
            atol=abstol_c,
        )

    def test_velocity_centre(self, hf):
        """
        Check that the velocity_centre function returns the expected velocity centre.
        """
        # default is minpot == 201. km/s
        assert_allclose_units(
            hf.velocity_centre,
            cosmo_array(
                [vcentre_1 + 1.0, vcentre_1 + 1.0, vcentre_1 + 1.0],
                u.km / u.s,
                comoving=True,
                cosmo_factor=cosmo_factor(a**0, 1.0),
            ),
            rtol=reltol_nd,
            atol=abstol_v,
        )


class TestHaloCataloguesMulti:
    def test_multi_flags(self, hf_multi):
        """
        Check that the multi-target nature of the cataloue is recognized.
        """
        assert hf_multi._multi_galaxy
        assert hf_multi._multi_galaxy_catalogue_mask is None
        assert hf_multi._multi_galaxy_index_mask is None
        assert hf_multi._multi_count == 2
        assert hf_multi.count == 2

    def test_mask_multi_galaxy(self, hf_multi):
        """
        Check that we can mask the catalogue to focus on one object, and unmask.
        """
        assert hf_multi._multi_galaxy_catalogue_mask is None
        assert hf_multi._multi_galaxy_index_mask is None
        assert hf_multi.count > 1
        assert hf_multi._region_centre.shape == (hf_multi.count, 3)
        assert hf_multi._region_aperture.shape == (hf_multi.count,)
        assert hf_multi.centre.shape == (hf_multi.count, 3)
        assert hf_multi.velocity_centre.shape == (hf_multi.count, 3)
        mask_index = 0
        hf_multi._mask_multi_galaxy(mask_index)
        if isinstance(hf_multi._multi_galaxy_index_mask, int):
            assert hf_multi._multi_galaxy_index_mask == mask_index
        elif isinstance(hf_multi._multi_galaxy_index_mask, slice):
            assert (
                hf_multi._multi_galaxy_index_mask == np.s_[mask_index : mask_index + 1]
            )
        if hf_multi._index_attr is not None:  # skip for Standalone
            assert (
                hf_multi._multi_galaxy_catalogue_mask
                == np.argsort(np.argsort(getattr(hf_multi, hf_multi._index_attr)))[
                    mask_index
                ]
            )
        assert hf_multi.count == 1
        assert hf_multi._region_centre.shape == (3,)
        if isinstance(hf_multi._multi_galaxy_index_mask, int):
            assert hf_multi.centre.shape == (3,)
            assert hf_multi.velocity_centre.shape == (3,)
        elif isinstance(hf_multi._multi_galaxy_index_mask, slice):
            assert hf_multi.centre.shape == (1, 3)
            assert hf_multi.velocity_centre.shape == (1, 3)
        assert hf_multi._region_aperture.shape == tuple()
        hf_multi._unmask_multi_galaxy()
        assert hf_multi._multi_galaxy_catalogue_mask is None
        assert hf_multi._multi_galaxy_index_mask is None
        assert hf_multi.count > 1
        assert hf_multi._region_centre.shape == (hf_multi.count, 3)
        assert hf_multi._region_aperture.shape == (hf_multi.count,)
        assert hf_multi.centre.shape == (hf_multi.count, 3)
        assert hf_multi.velocity_centre.shape == (hf_multi.count, 3)

    def test_get_spatial_mask(self, hf_multi, toysnap):
        """
        Check that we get spatial masks that we expect.
        """
        with pytest.raises(RuntimeError, match="not currently masked"):
            hf_multi._get_spatial_mask(toysnap_filename)
        hf_multi._mask_multi_galaxy(0)
        spatial_mask = hf_multi._get_spatial_mask(toysnap_filename)
        with h5py.File(toysnap_filename, "r") as snap:
            n_g_firstcell = snap["/Cells/Counts/PartType0"][0]
            n_dm_firstcell = snap["/Cells/Counts/PartType1"][0]
            n_s_firstcell = snap["/Cells/Counts/PartType4"][0]
            n_bh_firstcell = snap["/Cells/Counts/PartType5"][0]
        if hf_multi._user_spatial_offsets is None:
            # all hf's except Standalone
            assert np.array_equal(spatial_mask.gas, np.array([[0, n_g_firstcell]]))
            assert np.array_equal(
                spatial_mask.dark_matter, np.array([[0, n_dm_firstcell]])
            )
            assert np.array_equal(spatial_mask.stars, np.array([[0, n_s_firstcell]]))
            assert np.array_equal(
                spatial_mask.black_holes, np.array([[0, n_bh_firstcell]])
            )
        else:
            # this is Standalone
            assert np.array_equal(
                spatial_mask.gas,
                np.array([[0, n_g_firstcell], [n_g_firstcell, n_g_all]]),
            )
            assert np.array_equal(
                spatial_mask.dark_matter,
                np.array([[0, n_dm_firstcell], [n_dm_firstcell, n_dm_all]]),
            )
            assert np.array_equal(
                spatial_mask.stars,
                np.array([[0, n_s_firstcell], [n_s_firstcell, n_s_1 + n_s_2]]),
            )
            assert np.array_equal(
                spatial_mask.black_holes,
                np.array([[0, n_bh_firstcell], [n_bh_firstcell, n_bh_1 + n_bh_2]]),
            )

    def test_generate_extra_mask(self, hf_multi, toysnap_withfof):
        """
        Check that bound_only extra mask has the right shape.
        """
        hf_multi.extra_mask = "bound_only"
        hf_multi._mask_multi_galaxy(0)
        if hasattr(hf_multi, "soap_file"):
            from toysnap import soap_script
            import os

            os.system(
                f"python {soap_script} "
                f"--absolute-paths "
                f"'{toysnap_filename}' "
                f"'{toysoap_membership_filebase}."
                "{file_nr}.hdf5' "
                f"'{toysoap_virtual_snapshot_filename}' "
                "0"
            )
            sg = SWIFTGalaxy(toysoap_virtual_snapshot_filename, hf_multi)
        else:
            try:
                sg = SWIFTGalaxy(toysnap_filename, hf_multi)
            except NotImplementedError:
                # expected for Standalone
                return
        generated_extra_mask = sg._extra_mask
        expected_shape = dict()
        for particle_type in present_particle_types.values():
            with h5py.File(toysnap_filename, "r") as snap:
                expected_shape[particle_type] = snap[
                    "Cells/Counts/PartType"
                    f"{dict(gas=0, dark_matter=1, stars=4, black_holes=5)[particle_type]}"
                ][0]
        if hasattr(hf_multi, "_caesar") and hf_multi.group_type == "galaxy":
            expected_shape["dark_matter"] = None
        for particle_type in present_particle_types.values():
            if expected_shape[particle_type] is not None:
                assert (
                    getattr(generated_extra_mask, particle_type).shape
                    == expected_shape[particle_type]
                )
                assert (
                    getattr(generated_extra_mask, particle_type).sum()
                    == dict(
                        gas=n_g_1, dark_matter=n_dm_1, stars=n_s_1, black_holes=n_bh_1
                    )[particle_type]
                )

    def test_mask_index_list(self, hf_multi):
        """
        Check that we mask the list of indices.
        """
        if hf_multi._index_attr is None:
            # this is Standalone, nothing to do
            return
        # strip the leading underscore with [1:] to access the property
        assert getattr(hf_multi, hf_multi._index_attr[1:]) == [0, 1]
        hf_multi._mask_multi_galaxy(0)
        assert getattr(hf_multi, hf_multi._index_attr[1:]) == 0

    def test_masked_catalogue_matches(self, hf_multi):
        mask_index = 0
        init_args = {"extra_mask": hf_multi.extra_mask}
        if hf_multi.__class__ != Standalone:
            init_args[hf_multi._index_attr[1:]] = getattr(
                hf_multi, hf_multi._index_attr
            )[mask_index]
        else:
            init_args["centre"] = hf_multi._centre
            init_args["velocity_centre"] = hf_multi._velocity_centre
        if hasattr(hf_multi, "centre_type"):
            init_args["centre_type"] = hf_multi.centre_type
        if hasattr(hf_multi, "velocity_centre_type"):
            init_args["velocity_centre_type"] = hf_multi.velocity_centre_type
        if hf_multi.__class__ == Caesar:
            init_args["group_type"] = hf_multi.group_type
        init_args[
            ("custom_" if hf_multi.__class__ != Standalone else "") + "spatial_offsets"
        ] = hf_multi._user_spatial_offsets
        if hf_multi.__class__ == SOAP:
            init_args["soap_file"] = hf_multi.soap_file
        elif hf_multi.__class__ == Velociraptor:
            init_args["velociraptor_files"] = hf_multi.velociraptor_files
        elif hf_multi.__class__ == Caesar:
            init_args["caesar_file"] = hf_multi.caesar_file
        hf = hf_multi.__class__(**init_args)
        hf_multi._mask_multi_galaxy(mask_index)
        if hf_multi.__class__ == SOAP:
            assert (
                hf.bound_subhalo.enclose_radius == hf_multi.bound_subhalo.enclose_radius
            )
            assert (
                hf.bound_subhalo.enclose_radius.shape
                == hf_multi.bound_subhalo.enclose_radius.shape
            )
        elif hf_multi.__class__ == Velociraptor:
            assert hf_multi.masses.mass_200crit == hf.masses.mass_200crit
            assert hf_multi.masses.mass_200crit.shape == hf.masses.mass_200crit.shape
        elif hf_multi.__class__ == Caesar:
            assert hf_multi.masses["total"] == hf.masses["total"]
            assert hf_multi.masses["total"].shape == hf.masses["total"].shape
        elif hf_multi.__class__ == Standalone:
            pass  # has no catalogue to check
        else:
            raise NotImplementedError  # a new class we're not checking


class TestVelociraptor:
    def test_load(self, vr):
        """
        Check that the loading function is doing it's job.
        """
        # _load called during super().__init__
        assert vr._catalogue is not None

    @pytest.mark.parametrize(
        "centre_type, expected",
        (
            ("", centre_1),
            ("minpot", centre_1 + 0.001),
            ("mbp", centre_1 + 0.002),
            ("_gas", centre_1 + 0.003),
            ("_stars", centre_1 + 0.004),
        ),
    )
    def test_centre_types(self, vr, centre_type, expected):
        """
        Check that centres of each type retrieve expected values.
        """
        vr.centre_type = centre_type
        assert_allclose_units(
            vr.centre,
            cosmo_array(
                [expected, expected, expected],
                u.Mpc,
                comoving=True,
                cosmo_factor=cosmo_factor(a**1, 1.0),
            ),
            rtol=reltol_nd,
            atol=abstol_c,
        )

    @pytest.mark.parametrize(
        "centre_type, expected",
        (
            ("", vcentre_1),
            ("minpot", vcentre_1 + 1.0),
            ("mbp", vcentre_1 + 2.0),
            ("_gas", vcentre_1 + 3.0),
            ("_stars", vcentre_1 + 4.0),
        ),
    )
    def test_velocity_centre_types(self, vr, centre_type, expected):
        """
        Check that velocity centres of each type retrieve expected values.
        """
        vr.centre_type = centre_type
        assert_allclose_units(
            vr.velocity_centre,
            cosmo_array(
                [expected, expected, expected],
                u.km / u.s,
                comoving=True,
                cosmo_factor=cosmo_factor(a**0, 1.0),
            ),
            rtol=reltol_nd,
            atol=abstol_v,
        )

    def test_catalogue_exposed(self, vr):
        """
        Check that exposing the halo properties is working.
        """
        # pick one of the default attributes to check
        assert_allclose_units(
            vr.masses.mvir, 1.0e12 * u.Msun, rtol=reltol_nd, atol=abstol_m
        )

    def test_masking_catalogue(self, vr_multi):
        """
        Check that we can access unmasked and masked catalogue properties.
        """
        # pick one of the default attributes to check
        assert_allclose_units(
            vr_multi.masses.mvir,
            [1.0e12 * u.Msun, 1.1e12 * u.Msun],
            rtol=reltol_nd,
            atol=abstol_m,
        )
        vr_multi._mask_multi_galaxy(0)
        self.test_catalogue_exposed(vr_multi)


class TestVelociraptorWithSWIFTGalaxy:
    """
    Most interaction between the halo catalogue and swiftgalaxy.reader.SWIFTGalaxy
    is tested using the toysnap.ToyHF testing class (that inherits from
    swiftgalaxy.halo_catalogues._HaloCatalogue). Here we just want to test anything
    velociraptor-specific.
    """

    def test_catalogue_exposed(self, sg_vr):
        """
        Check that exposing the halo properties is working, through the
        SWIFTGalaxy object.
        """
        assert_allclose_units(
            sg_vr.halo_catalogue.masses.mvir,
            1.0e12 * u.Msun,
            rtol=reltol_nd,
            atol=abstol_m,
        )

    @pytest.mark.parametrize("particle_type", present_particle_types.values())
    def test_masks_compatible(self, sg_vr, particle_type):
        """
        Check that the bound_only default mask works with the spatial mask,
        giving the expected shapes for arrays.
        """
        assert (
            getattr(sg_vr, particle_type).masses.size
            == dict(gas=n_g_1, dark_matter=n_dm_1, stars=n_s_1, black_holes=n_bh_1)[
                particle_type
            ]
        )

    def test_with_swiftgalaxies(self, sgs_vr):
        for sg_from_sgs in sgs_vr:
            sg = SWIFTGalaxy(
                sg_from_sgs.snapshot_filename,
                Velociraptor(
                    velociraptor_files=sg_from_sgs.halo_catalogue.velociraptor_files,
                    halo_index=sg_from_sgs.halo_catalogue.halo_index,
                ),
            )
            for ptype in present_particle_types.values():
                assert np.all(
                    getattr(sg_from_sgs._extra_mask, ptype)
                    == getattr(sg._extra_mask, ptype)
                )


class TestCaesar:
    def test_load(self, caesar):
        """
        Check that the loading function is doing it's job.
        """
        # _load called during super().__init__
        pass  # Caesar has nothing to do in _load

    @pytest.mark.parametrize(
        "centre_type, expected", (("", centre_1), ("minpot", centre_1 + 0.001))
    )
    def test_centre_types(self, caesar, centre_type, expected):
        """
        Check that centres of each type retrieve expected values.
        """
        caesar.centre_type = centre_type
        assert_allclose_units(
            caesar.centre,
            cosmo_array(
                [expected, expected, expected],
                u.Mpc,
                comoving=True,
                cosmo_factor=cosmo_factor(a**1, 1.0),
            ),
            rtol=reltol_nd,
            atol=abstol_c,
        )

    @pytest.mark.parametrize(
        "centre_type, expected", (("", vcentre_1), ("minpot", vcentre_1 + 1.0))
    )
    def test_vcentre_types(self, caesar, centre_type, expected):
        """
        Check that velocity centres of each type retrieve expected values.
        """
        caesar.centre_type = centre_type
        assert_allclose_units(
            caesar.velocity_centre,
            cosmo_array(
                [expected, expected, expected],
                u.km / u.s,
                comoving=True,
                cosmo_factor=cosmo_factor(a**0, 1.0),
            ),
            rtol=reltol_nd,
            atol=abstol_v,
        )

    def test_catalogue_exposed(self, caesar):
        """
        Check that exposing the halo properties is working.
        """
        # pick one of the default attributes to check
        if hasattr(caesar, "virial_quantities"):
            assert_allclose_units(
                caesar.virial_quantities["m200c"],
                1.0e12 * u.Msun,
                rtol=reltol_nd,
                atol=abstol_m,
            )
        elif hasattr(caesar, "masses"):
            assert_allclose_units(
                caesar.masses["total"],
                n_g_1 * m_g + n_s_1 * m_s + n_bh_1 * m_bh,
                rtol=reltol_nd,
                atol=abstol_m,
            )
        else:
            raise AttributeError

    def test_masking_catalogue(self, caesar_multi):
        """
        Check that we can access unmasked and masked catalogue properties.
        """
        # pick one of the default attributes to check
        if hasattr(caesar_multi, "virial_quantities"):
            assert_allclose_units(
                [c["m200c"] for c in caesar_multi.virial_quantities],
                [1.0e12 * u.Msun, 2.0e12 * u.Msun],
                rtol=reltol_nd,
                atol=abstol_m,
            )
        elif hasattr(caesar_multi, "masses"):
            assert_allclose_units(
                [c["total"] for c in caesar_multi.masses],
                [
                    n_g_1 * m_g + n_s_1 * m_s + n_bh_1 * m_bh,
                    n_g_2 * m_g + n_s_2 * m_s + n_bh_2 * m_bh,
                ],
                rtol=reltol_nd,
                atol=abstol_m,
            )
        else:
            raise AttributeError
        caesar_multi._mask_multi_galaxy(0)
        self.test_catalogue_exposed(caesar_multi)

    def test_spatial_mask_applied(self, caesar, toysnap):
        """
        Check that we get the expected number of particles when only the spatial mask is
        applied.
        """
        caesar.extra_mask = None  # apply only the spatial mask
        sg = SWIFTGalaxy(toysnap_filename, caesar)
        for particle_type in present_particle_types.values():
            assert (
                getattr(sg, particle_type).masses.size
                == dict(
                    gas=n_g_b // 2 + n_g_1,
                    dark_matter=n_dm_b // 2 + n_dm_1,
                    stars=n_s_1,
                    black_holes=n_bh_1,
                )[particle_type]
            )


class TestCaesarWithSWIFTGalaxy:
    """
    Most interaction between the halo catalogue and swiftgalaxy.reader.SWIFTGalaxy
    is tested using the toysnap.ToyHF testing class (that inherits from
    swiftgalaxy.halo_catalogues._HaloCatalogue). Here we just want to test anything
    caesar-specific.
    """

    def test_catalogue_exposed(self, sg_caesar):
        """
        Check that exposing the halo properties is working, through the
        SWIFTGalaxy object.
        """
        if hasattr(sg_caesar.halo_catalogue, "virial_quantities"):
            assert_allclose_units(
                sg_caesar.halo_catalogue.virial_quantities["m200c"],
                1.0e12 * u.Msun,
                rtol=reltol_nd,
                atol=abstol_m,
            )
        elif hasattr(sg_caesar.halo_catalogue, "masses"):
            assert_allclose_units(
                sg_caesar.halo_catalogue.masses["total"],
                n_g_1 * m_g + n_s_1 * m_s + n_bh_1 * m_bh,
                rtol=reltol_nd,
                atol=abstol_m,
            )
        else:
            raise AttributeError

    @pytest.mark.parametrize("particle_type", present_particle_types.values())
    def test_masks_compatible(self, sg_caesar, particle_type):
        """
        Check that the bound_only default mask works with the spatial mask,
        giving the expected shapes for arrays.
        """
        expected_dm = 0 if sg_caesar.halo_catalogue.group_type == "galaxy" else n_dm_1
        assert (
            getattr(sg_caesar, particle_type).masses.size
            == dict(
                gas=n_g_1, dark_matter=expected_dm, stars=n_s_1, black_holes=n_bh_1
            )[particle_type]
        )

    def test_with_swiftgalaxies(self, sgs_caesar):
        for sg_from_sgs in sgs_caesar:
            sg = SWIFTGalaxy(
                sg_from_sgs.snapshot_filename,
                Caesar(
                    caesar_file=sg_from_sgs.halo_catalogue.caesar_file,
                    group_type=sg_from_sgs.halo_catalogue.group_type,
                    group_index=sg_from_sgs.halo_catalogue.group_index,
                ),
            )
            for ptype in present_particle_types.values():
                if isinstance(getattr(sg_from_sgs._extra_mask, ptype), slice):
                    assert getattr(sg_from_sgs._extra_mask, ptype) == getattr(
                        sg._extra_mask, ptype
                    )
                else:
                    assert np.all(
                        getattr(sg_from_sgs._extra_mask, ptype)
                        == getattr(sg._extra_mask, ptype)
                    )


class TestStandalone:
    def test_spatial_mask_applied(self, sa, toysnap):
        """
        Check that we get the expected number of particles when only the spatial mask is
        applied.
        """
        sa.extra_mask = None  # apply only the spatial mask
        sg = SWIFTGalaxy(toysnap_filename, sa)
        for particle_type in present_particle_types.values():
            assert (
                getattr(sg, particle_type).masses.size
                == dict(
                    gas=n_g_b // 2 + n_g_1,
                    dark_matter=n_dm_b // 2 + n_dm_1,
                    stars=n_s_1,
                    black_holes=n_bh_1,
                )[particle_type]
            )


class TestSOAP:
    def test_load(self, soap):
        """
        Check that the loading function is doing it's job.
        """
        # _load called during super().__init__
        assert soap._catalogue is not None

    @pytest.mark.parametrize(
        "centre_type, expected",
        (
            ("bound_subhalo.centre_of_mass", centre_1 + 0.001),
            ("exclusive_sphere_100kpc.centre_of_mass", centre_1 + 0.003),
            ("inclusive_sphere_100kpc.centre_of_mass", centre_1 + 0.011),
            ("input_halos_fof.centres", centre_1),
            ("input_halos.halo_centre", centre_1 + 0.001),
            ("projected_aperture_50kpc_projx.centre_of_mass", centre_1 + 0.027),
            ("spherical_overdensity_200_crit.centre_of_mass", centre_1 + 0.032),
            ("spherical_overdensity_bn98.centre_of_mass", centre_1 + 0.038),
        ),
    )
    def test_centre_types(self, soap, centre_type, expected):
        """
        Check that centres of sample types retrieve expected values.
        """
        soap.centre_type = centre_type
        assert_allclose_units(
            soap.centre,
            cosmo_array(
                [expected, expected, expected],
                u.Mpc,
                comoving=True,
                cosmo_factor=cosmo_factor(a**1, 1.0),
            ),
            rtol=reltol_nd,
            atol=abstol_c,
        )

    @pytest.mark.parametrize(
        "velocity_centre_type, expected",
        (
            ("bound_subhalo.centre_of_mass_velocity", vcentre_1 + 1),
            ("exclusive_sphere_100kpc.centre_of_mass_velocity", vcentre_1 + 3),
            ("inclusive_sphere_100kpc.centre_of_mass_velocity", vcentre_1 + 11),
            ("projected_aperture_50kpc_projx.centre_of_mass_velocity", vcentre_1 + 27),
            ("spherical_overdensity_200_crit.centre_of_mass_velocity", vcentre_1 + 32),
            ("spherical_overdensity_bn98.centre_of_mass_velocity", vcentre_1 + 38),
        ),
    )
    def test_velocity_centre_types(self, soap, velocity_centre_type, expected):
        """
        Check that velocity centres of sample types retrieve expected values.
        """
        soap.velocity_centre_type = velocity_centre_type
        assert_allclose_units(
            soap.velocity_centre,
            cosmo_array(
                [expected, expected, expected],
                u.km / u.s,
                comoving=True,
                cosmo_factor=cosmo_factor(a**0, 1.0),
            ),
            rtol=reltol_nd,
            atol=abstol_v,
        )

    def test_catalogue_exposed(self, soap):
        """
        Check that exposing the halo properties is working.
        """
        # pick a couple of attributes to check
        assert_allclose_units(
            soap.input_halos_hbtplus.host_fofid, cosmo_array([1], comoving=False)
        )
        assert_allclose_units(
            soap.bound_subhalo.centre_of_mass,
            cosmo_array(
                [[centre_1 + 0.001, centre_1 + 0.001, centre_1 + 0.001]],
                u.Mpc,
                comoving=False,
                cosmo_factor=cosmo_factor(a**1, 1.0),
            ),
            rtol=reltol_nd,
            atol=abstol_c,
        )

    def test_masking_catalogue(self, soap_multi):
        """
        Check that we can access unmasked and masked catalogue properties.
        """
        # pick a couple of attributes to check
        assert_allclose_units(
            soap_multi.input_halos_hbtplus.host_fofid,
            cosmo_array([1, 2], comoving=False),
        )
        assert_allclose_units(
            soap_multi.bound_subhalo.centre_of_mass,
            cosmo_array(
                [
                    [centre_1 + 0.001, centre_1 + 0.001, centre_1 + 0.001],
                    [centre_2 + 0.001, centre_2 + 0.001, centre_2 + 0.001],
                ],
                u.Mpc,
                comoving=False,
                cosmo_factor=cosmo_factor(a**1, 1.0),
            ),
            rtol=reltol_nd,
            atol=abstol_c,
        )
        soap_multi._mask_multi_galaxy(0)
        self.test_catalogue_exposed(soap_multi)


class TestSOAPWithSWIFTGalaxy:
    """
    Most interaction between the halo catalogue and swiftgalaxy.reader.SWIFTGalaxy
    is tested using the toysnap.ToyHF testing class (that inherits from
    swiftgalaxy.halo_catalogues._HaloCatalogue). Here we just want to test anything
    soap-specific.
    """

    def test_catalogue_exposed(self, sg_soap):
        """
        Check that exposing the halo properties is working, through the
        SWIFTGalaxy object.
        """
        # pick a couple of attributes to check
        assert_allclose_units(
            sg_soap.halo_catalogue.input_halos_hbtplus.host_fofid,
            cosmo_array([1], comoving=False),
        )
        assert_allclose_units(
            sg_soap.halo_catalogue.bound_subhalo.centre_of_mass,
            cosmo_array(
                [[centre_1 + 0.001, centre_1 + 0.001, centre_1 + 0.001]],
                u.Mpc,
                comoving=False,
                cosmo_factor=cosmo_factor(a**1, 1.0),
            ),
            rtol=reltol_nd,
            atol=abstol_c,
        )

    @pytest.mark.parametrize("particle_type", present_particle_types.values())
    def test_masks_compatible(self, sg_soap, particle_type):
        """
        Check that the bound_only default mask works with the spatial mask,
        giving the expected shapes for arrays.
        """
        assert (
            getattr(sg_soap, particle_type).masses.size
            == dict(gas=n_g_1, dark_matter=n_dm_1, stars=n_s_1, black_holes=n_bh_1)[
                particle_type
            ]
        )

    def test_with_swiftgalaxies(self, sgs_soap):
        for sg_from_sgs in sgs_soap:
            sg = SWIFTGalaxy(
                sg_from_sgs.snapshot_filename,
                SOAP(
                    soap_file=sg_from_sgs.halo_catalogue.soap_file,
                    soap_index=sg_from_sgs.halo_catalogue.soap_index,
                ),
            )
            for ptype in present_particle_types.values():
                assert np.all(
                    getattr(sg_from_sgs._extra_mask, ptype)
                    == getattr(sg._extra_mask, ptype)
                )
